import re
import typing
from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Union

import pandas
import pendulum
from google.protobuf import duration_pb2

from tecton_core import conf
from tecton_core import data_types
from tecton_core import errors
from tecton_core import specs
from tecton_core.errors import INVALID_SPINE_SQL
from tecton_core.errors import START_TIME_NOT_BEFORE_END_TIME
from tecton_core.errors import TectonSnowflakeNotImplementedError
from tecton_core.feature_definition_wrapper import FeatureDefinitionWrapper
from tecton_core.feature_set_config import FeatureSetConfig
from tecton_core.materialization_context import BoundMaterializationContext
from tecton_core.materialization_context import MaterializationContext
from tecton_core.query_consts import udf_internal
from tecton_core.time_utils import convert_pandas_df_for_snowflake_upload
from tecton_proto.common import aggregation_function__client_pb2 as afpb
from tecton_proto.data import feature_view__client_pb2 as feature_view_pb2
from tecton_snowflake.pipeline_helper import pipeline_to_df_with_input
from tecton_snowflake.pipeline_helper import pipeline_to_sql_string
from tecton_snowflake.templates_utils import load_template
from tecton_snowflake.utils import format_sql
from tecton_snowflake.utils import generate_random_name


if typing.TYPE_CHECKING:
    import snowflake.connector
    import snowflake.snowpark


TEMP_SPINE_TABLE_NAME_FROM_DF = "_TEMP_SPINE_TABLE_FROM_DF"
TEMP_SPINE_TABLE_NAME_PREFIX = "_TT_SPINE_TABLE"

FULL_AGGREGATION_TEMPLATE = load_template("run_full_aggregation.sql")
HISTORICAL_FEATURES_TEMPLATE = load_template("historical_features.sql")
MATERIALIZATION_TILE_TEMPLATE = load_template("materialization_tile.sql")
MATERIALIZED_FEATURE_VIEW_TEMPLATE = load_template("materialized_feature_view.sql")
ONLINE_STORE_COPIER_TEMPLATE = load_template("online_store_copier.sql")
PARTIAL_AGGREGATION_TEMPLATE = load_template("run_partial_aggregation.sql")
TIME_LIMIT_TEMPLATE = load_template("time_limit.sql")
DELETE_STAGED_FILES_TEMPLATE = load_template("delete_staged_files.sql")
CREATE_TEMP_TABLE_BFV_TEMPLATE = load_template("create_temp_table_for_bfv.sql")
CREATE_TEMP_TABLE_BWAFV_TEMPLATE = load_template("create_temp_table_for_bwafv.sql")
DELETE_ORPHANED_SCHEMAS_TEMPLATE = load_template("delete_orphaned_schemas.sql")

# TODO(TEC-6204): Last and LastN are not currently supported
#
# Map of proto function type -> set of (output column prefix, snowflake function name)
AGGREGATION_PLANS = {
    afpb.AGGREGATION_FUNCTION_SUM: lambda: {("SUM", "SUM")},
    afpb.AGGREGATION_FUNCTION_MIN: lambda: {("MIN", "MIN")},
    afpb.AGGREGATION_FUNCTION_MAX: lambda: {("MAX", "MAX")},
    afpb.AGGREGATION_FUNCTION_COUNT: lambda: {("COUNT", "COUNT")},
    afpb.AGGREGATION_FUNCTION_MEAN: lambda: {("COUNT", "COUNT"), ("MEAN", "AVG")},
    # sample variation equation: (Σ(x^2) - (Σ(x)^2)/N)/N-1
    afpb.AGGREGATION_FUNCTION_VAR_SAMP: lambda: {
        ("SUM", "SUM"),
        ("COUNT", "COUNT"),
        ("SUM_OF_SQUARES", "SUM_OF_SQUARES"),
    },
    # population variation equation: Σ(x^2)/n - μ^2
    afpb.AGGREGATION_FUNCTION_VAR_POP: lambda: {
        ("SUM", "SUM"),
        ("COUNT", "COUNT"),
        ("SUM_OF_SQUARES", "SUM_OF_SQUARES"),
    },
    # sample standard deviation equation: √ ((Σ(x^2) - (Σ(x)^2)/N)/N-1)
    afpb.AGGREGATION_FUNCTION_STDDEV_SAMP: lambda: {
        ("SUM", "SUM"),
        ("COUNT", "COUNT"),
        ("SUM_OF_SQUARES", "SUM_OF_SQUARES"),
    },
    # population standard deviation equation: √ (Σ(x^2)/n - μ^2)
    afpb.AGGREGATION_FUNCTION_STDDEV_POP: lambda: {
        ("SUM", "SUM"),
        ("COUNT", "COUNT"),
        ("SUM_OF_SQUARES", "SUM_OF_SQUARES"),
    },
    afpb.AGGREGATION_FUNCTION_LAST_NON_DISTINCT_N: lambda n, is_continuous: {
        ("LAST_NON_DISTINCT_N" + (str(n) if not is_continuous else ""), n),
    },
    afpb.AGGREGATION_FUNCTION_FIRST_NON_DISTINCT_N: lambda n, is_continuous: {
        ("FIRST_NON_DISTINCT_N" + (str(n) if not is_continuous else ""), n),
    },
}


@dataclass
class _FeatureSetItemInput:
    """A simplified version of FeatureSetItem which is passed to the SQL template."""

    name: str
    namespace: str
    timestamp_key: str
    join_keys: Dict[str, str]
    features: List[str]
    sql: str
    aggregation: Optional[feature_view_pb2.TrailingTimeWindowAggregation]
    ttl_seconds: Optional[int]
    append_prefix: bool
    batch_schedule: int
    data_delay: int


def get_historical_features_sql(
    spine_sql: Optional[str],
    feature_set_config: FeatureSetConfig,
    timestamp_key: Optional[str],
    from_source: Optional[bool],
    include_feature_view_timestamp_columns: bool = False,
    start_time: Optional[datetime] = None,
    end_time: Optional[datetime] = None,
    session: "snowflake.snowpark.Session" = None,
    append_prefix: bool = True,  # Whether to append the prefix to the feature column name
) -> List[str]:
    suffix = generate_random_name()
    spine_table_name = f"{TEMP_SPINE_TABLE_NAME_PREFIX}_{suffix}"
    # Whether to register temp views with the session, or use a giant query
    use_short_sql = conf.get_bool("SNOWFLAKE_SHORT_SQL_ENABLED")
    use_temp_table = conf.get_bool("SNOWFLAKE_TEMP_TABLE_ENABLED")
    feature_set_items = feature_set_config.definitions_and_configs
    input_items = []
    if spine_sql is None:
        # Only feature view is supported when the spine is not provided.
        # Feature service should always provide the spine.
        # SDK methods should never fail this check
        assert len(feature_set_items) == 1

    # Get a list of all the join keys in the spine.
    spine_keys = {}
    for item in feature_set_items:
        fd = item.feature_definition

        if not fd.is_realtime:
            join_keys = dict(item.join_keys)
            spine_keys.update(join_keys)

    for item in feature_set_items:
        fd = item.feature_definition
        # Change the feature view name if it's for internal udf use.
        is_internal_udf_feature = item.namespace.startswith(udf_internal())
        if is_internal_udf_feature:
            name = item.namespace.upper()
        else:
            name = fd.name

        if not fd.is_realtime:
            join_keys = dict(item.join_keys)
            features = [
                col_name
                for col_name in fd.view_schema.column_names()
                if col_name not in ([*list(join_keys.keys()), fd.timestamp_key])
            ]
            if len(fd.online_serving_index.join_keys) != len(fd.join_keys):
                msg = "Wildcard is not supported for Snowflake"
                raise TectonSnowflakeNotImplementedError(msg)
            if fd.is_incremental_backfill and from_source:
                raise errors.FV_BFC_SINGLE_FROM_SOURCE
            if start_time is None or (
                fd.feature_start_timestamp is not None and start_time < fd.feature_start_timestamp
            ):
                raw_data_start_time = fd.feature_start_timestamp
            else:
                raw_data_start_time = start_time

            raw_data_end_time = end_time
            if fd.is_temporal_aggregate and raw_data_start_time is not None:
                # Account for final aggregation needing aggregation window prior to earliest timestamp
                raw_data_start_time = raw_data_start_time + fd.earliest_window_start

            sql_str = generate_run_batch_sql(
                feature_definition=fd,
                feature_start_time=raw_data_start_time,
                feature_end_time=raw_data_end_time,
                # If the spine sql has undetermined results(LIMIT X etc.), we don't want
                # to run it twice. So we use the subquery we defined here directly.
                spine=spine_table_name if spine_sql is not None else None,
                spine_timestamp_key=timestamp_key,
                session=session,
                from_source=from_source,
                data_delay=fd.max_source_data_delay,
            )
            input_items.append(
                _FeatureSetItemInput(
                    name=name,
                    namespace=item.namespace or name,
                    timestamp_key=fd.timestamp_key,
                    join_keys=join_keys,
                    features=features,
                    sql=sql_str,
                    aggregation=(fd.trailing_time_window_aggregation() if fd.is_temporal_aggregate else None),
                    ttl_seconds=(int(fd.serving_ttl.total_seconds()) if fd.is_temporal else None),
                    append_prefix=append_prefix or is_internal_udf_feature,
                    batch_schedule=0
                    if fd.min_scheduling_interval is None
                    else int(fd.min_scheduling_interval.total_seconds()),
                    data_delay=fd.max_source_data_delay.total_seconds(),
                )
            )
    queries = []

    if spine_sql is not None:
        if use_short_sql or use_temp_table:
            view_or_table = "TABLE" if use_temp_table else "VIEW"
            queries.append(f"CREATE TEMPORARY {view_or_table} {spine_table_name} AS ({spine_sql})")
            for item in input_items:
                if item.aggregation is not None:
                    queries.append(
                        format_sql(
                            CREATE_TEMP_TABLE_BWAFV_TEMPLATE.render(
                                item=item,
                                spine_table_name=spine_table_name,
                                spine_keys=list(spine_keys),
                                spine_timestamp_key=timestamp_key,
                                suffix=suffix,
                                view_or_table=view_or_table,
                            )
                        )
                    )
                else:
                    queries.append(
                        format_sql(
                            CREATE_TEMP_TABLE_BFV_TEMPLATE.render(
                                item=item,
                                spine_table_name=spine_table_name,
                                spine_keys=list(spine_keys),
                                spine_timestamp_key=timestamp_key,
                                include_feature_view_timestamp_columns=include_feature_view_timestamp_columns,
                                suffix=suffix,
                                view_or_table=view_or_table,
                            )
                        )
                    )
        sql_str = HISTORICAL_FEATURES_TEMPLATE.render(
            feature_set_items=input_items,
            spine_timestamp_key=timestamp_key,
            spine_sql=spine_sql,
            include_feature_view_timestamp_columns=include_feature_view_timestamp_columns,
            spine_keys=list(spine_keys),
            use_temp_tables=use_short_sql or use_temp_table,
            spine_table_name=spine_table_name,
            suffix=suffix,
        )
    if start_time is not None or end_time is not None:
        timestamp = timestamp_key if spine_sql is not None else fd.timestamp_key
        sql_str = TIME_LIMIT_TEMPLATE.render(
            source=sql_str, timestamp_key=timestamp, start_time=start_time, end_time=end_time
        )
    queries.append(format_sql(sql_str))
    if conf.get_bool("SNOWFLAKE_DEBUG"):
        print(f"Generated {len(queries)} Snowflake SQL for get_historical_features, this does not include ODFV:")
        for query in queries:
            print(query)
    return queries


def get_historical_features(
    spine: Optional[Union[pandas.DataFrame, str]],
    connection: "snowflake.connector.Connection",
    feature_set_config: FeatureSetConfig,
    timestamp_key: Optional[str],
    from_source: Optional[bool],
    include_feature_view_timestamp_columns: bool = False,
    start_time: Optional[datetime] = None,
    end_time: Optional[datetime] = None,
    append_prefix: bool = True,  # Whether to append the prefix to the feature column name
) -> pandas.DataFrame:
    cur = connection.cursor()
    spine_sql = None
    if isinstance(spine, str):
        spine_sql = spine
    elif isinstance(spine, pandas.DataFrame):
        spine_sql = generate_sql_table_from_pandas_df(
            df=spine, table_name=TEMP_SPINE_TABLE_NAME_FROM_DF, connection=connection
        )

    sql_strs = get_historical_features_sql(
        spine_sql=spine_sql,
        feature_set_config=feature_set_config,
        timestamp_key=timestamp_key,
        include_feature_view_timestamp_columns=include_feature_view_timestamp_columns,
        start_time=start_time,
        end_time=end_time,
        append_prefix=append_prefix,
        from_source=from_source,
    )
    for sql_str in sql_strs:
        cur.execute(sql_str, _statement_params={"SF_PARTNER": "tecton-ai"})
    return cur.fetch_pandas_all()


def generate_sql_table_from_pandas_df(
    df: pandas.DataFrame,
    table_name: str,
    session: "snowflake.snowpark.Session" = None,
    connection: "snowflake.connector.Connection" = None,
) -> str:
    """Generate a TABLE from pandas.DataFrame. Returns the sql query to select * from the table"""
    if session is None and connection is None:
        msg = "Either session or connection must be provided"
        raise ValueError(msg)

    if session is not None:
        session.sql(f"DROP TABLE IF EXISTS {table_name}").collect(statement_params={"SF_PARTNER": "tecton-ai"})
        convert_pandas_df_for_snowflake_upload(df)
        session.write_pandas(df, table_name, auto_create_table=True, table_type="temporary", use_logical_type=True)
        return f"SELECT * FROM {table_name}"

    if connection is not None:
        from snowflake.connector.pandas_tools import write_pandas

        # Get the SQL that would be generated by the create table statement
        create_table_sql = pandas.io.sql.get_schema(df, table_name)

        # Replace the `CREATE TABLE` with `CREATE OR REPLACE TEMPORARY TABLE`
        create_tmp_table_sql = re.sub("^(CREATE TABLE)?", "CREATE OR REPLACE TEMPORARY TABLE", create_table_sql)
        cur = connection.cursor()
        cur.execute(create_tmp_table_sql, _statement_params={"SF_PARTNER": "tecton-ai"})
        write_pandas(conn=connection, df=df, table_name=table_name, use_logical_type=True)
        return f"SELECT * FROM {table_name}"


def validate_spine_dataframe(
    spine_df: "snowflake.snowpark.DataFrame",
    timestamp_key: str,
    join_keys: List[str],
    request_context_keys: List[str] = [],
):
    from snowflake.snowpark import types

    if timestamp_key not in spine_df.columns:
        msg = f"Expected to find '{timestamp_key}' among available spine columns: '{', '.join(spine_df.columns)}'."
        raise errors.TectonValidationError(msg)
    for field in spine_df.schema.fields:
        if field.name == timestamp_key and not isinstance(field.datatype, types.TimestampType):
            msg = f"Invalid type of timestamp_key column in the given spine. Expected Timestamp, got {field.datatype}"
            raise errors.TectonValidationError(msg)
    for key in join_keys + request_context_keys:
        if key not in spine_df.columns:
            msg = f"Expected to find '{key}' among available spine columns: '{', '.join(spine_df.columns)}'."
            raise errors.TectonValidationError(msg)


def get_historical_features_with_snowpark(
    spine: Union[pandas.DataFrame, str, "snowflake.snowpark.DataFrame"],
    session: "snowflake.snowpark.Session",
    feature_set_config: FeatureSetConfig,
    timestamp_key: Optional[str],
    from_source: Optional[bool],
    include_feature_view_timestamp_columns: bool = False,
    start_time: Optional[datetime] = None,
    end_time: Optional[datetime] = None,
    entities: "snowflake.snowpark.DataFrame" = None,
    append_prefix: bool = True,  # Whether to append the prefix to the feature column name
) -> "snowflake.snowpark.DataFrame":
    from snowflake.snowpark import DataFrame
    from snowflake.snowpark import types

    spine_sql = None
    if isinstance(spine, str):
        spine_sql = spine
    elif isinstance(spine, DataFrame):
        spine.write.save_as_table(TEMP_SPINE_TABLE_NAME_FROM_DF, mode="overwrite", table_type="temporary")
        spine_sql = f"SELECT * FROM {TEMP_SPINE_TABLE_NAME_FROM_DF}"
    elif isinstance(spine, pandas.DataFrame):
        spine_sql = generate_sql_table_from_pandas_df(
            df=spine, table_name=TEMP_SPINE_TABLE_NAME_FROM_DF, session=session
        )

    if spine is not None:
        # validate spine sql
        try:
            spine_df = session.sql(spine_sql)
            spine_schema = spine_df.schema
        except Exception as e:
            raise INVALID_SPINE_SQL(e)
        if timestamp_key is None:
            schema = spine_schema
            timestamp_cols = [field.name for field in schema.fields if isinstance(field.datatype, types.TimestampType)]

            if len(timestamp_cols) > 1 or len(timestamp_cols) == 0:
                msg = f"Could not infer timestamp keys from {schema}; please specify explicitly"
                raise errors.TectonValidationError(msg)
            timestamp_key = timestamp_cols[0]

        join_keys = [join_key for fd in feature_set_config.feature_definitions for join_key in fd.join_keys]
        request_context_keys = [key for fd in feature_set_config.feature_definitions for key in fd.request_context_keys]
        validate_spine_dataframe(
            spine_df, timestamp_key, join_keys=join_keys, request_context_keys=request_context_keys
        )

    sql_strs = get_historical_features_sql(
        spine_sql=spine_sql,
        feature_set_config=feature_set_config,
        timestamp_key=timestamp_key,
        include_feature_view_timestamp_columns=include_feature_view_timestamp_columns,
        start_time=start_time,
        end_time=end_time,
        session=session,
        append_prefix=append_prefix,
        from_source=from_source,
    )
    for sql_str in sql_strs[:-1]:
        session.sql(sql_str).collect(statement_params={"SF_PARTNER": "tecton-ai"})
    output_df = session.sql(sql_strs[-1])

    # Apply ODFV to the spine.
    for item in feature_set_config.definitions_and_configs:
        fd = item.feature_definition
        if fd.is_realtime:
            schema_dict = fd.view_schema.to_dict()
            output_df = pipeline_to_df_with_input(
                session=session,
                input_df=output_df,
                pipeline=fd.pipeline,
                transformations=fd.transformations,
                output_schema=schema_dict,
                name=fd.name,
                fv_id=fd.id,
                namespace=item.namespace or fd.name,
                append_prefix=append_prefix,
            )
    columns_to_drop = [column for column in output_df.columns if "_UDF_INTERNAL_" in column]
    if len(columns_to_drop) > 0:
        output_df = output_df.drop(*columns_to_drop)
    if entities is not None:
        column_names = [field.name for field in entities.schema.fields]
        # Do an inner join on the entities to filter out rows that don't have a matching entity
        output_df = entities.join(right=output_df, using_columns=column_names, join_type="inner")
    return output_df


def generate_run_batch_sql(
    feature_definition: FeatureDefinitionWrapper,
    from_source: Optional[bool],
    # start is inclusive and end is exclusive
    feature_start_time: Optional[datetime] = None,
    feature_end_time: Optional[datetime] = None,
    aggregation_level: str = "full",
    # If spine is provided, it will be used to join with the output results.
    # Currently only work with full aggregation.
    spine: Optional[str] = None,
    spine_timestamp_key: Optional[str] = None,
    spine_keys: Optional[List[str]] = None,
    mock_sql_inputs: Optional[Dict[str, str]] = None,
    materialization_context: Optional[MaterializationContext] = None,
    session: "snowflake.snowpark.Session" = None,
    data_delay: Optional[pendulum.Duration] = None,
) -> str:
    # Set a default materilization_context if not provided.
    # This is following the same logic as spark.
    if materialization_context is None:
        materialization_feature_start_time = feature_start_time or pendulum.from_timestamp(0, pendulum.tz.UTC)
        materialization_feature_end_time = feature_end_time or pendulum.datetime(2100, 1, 1)
        if not materialization_feature_start_time < materialization_feature_end_time:
            raise START_TIME_NOT_BEFORE_END_TIME(materialization_feature_start_time, materialization_feature_end_time)

        materialization_context = BoundMaterializationContext._create_internal(
            materialization_feature_start_time,
            materialization_feature_end_time,
            feature_definition.fv_spec.batch_schedule,
        )

    if from_source is None:
        from_source = not feature_definition.materialization_enabled or not feature_definition.writes_to_offline_store

    if from_source is False:
        if not feature_definition.materialization_enabled or not feature_definition.writes_to_offline_store:
            raise errors.FV_NEEDS_TO_BE_MATERIALIZED(feature_definition.name)

    if from_source:
        pipeline_sql = pipeline_to_sql_string(
            pipeline=feature_definition.pipeline,
            data_sources=feature_definition.data_sources,
            transformations=feature_definition.transformations,
            materialization_context=materialization_context,
            mock_sql_inputs=mock_sql_inputs,
            session=session,
        )
        materialized_sql = get_materialization_query(
            feature_definition=feature_definition,
            feature_start_time=feature_start_time,
            feature_end_time=feature_end_time,
            source=pipeline_sql,
            offline_materialization=True,
        )
    else:
        materialized_sql = TIME_LIMIT_TEMPLATE.render(
            source=f"SELECT * FROM {feature_definition.fv_spec.snowflake_view_name}",
            timestamp_key=feature_definition.time_key,
            start_time=feature_start_time,
            end_time=feature_end_time,
        )
    if feature_definition.is_temporal_aggregate:
        if aggregation_level == "full":
            aggregated_sql_str = FULL_AGGREGATION_TEMPLATE.render(
                source=materialized_sql,
                join_keys=feature_definition.join_keys,
                aggregation=feature_definition.trailing_time_window_aggregation(),
                timestamp_key=feature_definition.time_key,
                name=feature_definition.name,
                spine=spine,
                spine_timestamp_key=spine_timestamp_key,
                spine_keys=spine_keys,
                batch_schedule=int(materialization_context.batch_schedule.total_seconds()),
                data_delay=int(data_delay.total_seconds()) if data_delay else 0,
                continuous=feature_definition.is_continuous,
                offline_materialization=True,
                is_continuous=feature_definition.is_continuous,
            )
            return format_sql(aggregated_sql_str)
        elif aggregation_level == "partial":
            # Rename the output columns, and add tile start/end time columns
            partial_aggregated_sql_str = PARTIAL_AGGREGATION_TEMPLATE.render(
                source=materialized_sql,
                join_keys=feature_definition.join_keys,
                aggregations=_get_feature_view_aggregations(feature_definition),
                slide_interval=feature_definition.aggregate_slide_interval,
                slide_interval_string=feature_definition.get_aggregate_slide_interval_string,
                timestamp_key=feature_definition.time_key,
            )
            return format_sql(partial_aggregated_sql_str)
        elif aggregation_level == "disabled":
            sql_str = TIME_LIMIT_TEMPLATE.render(
                source=pipeline_sql,
                timestamp_key=feature_definition.time_key,
                start_time=feature_start_time,
                end_time=feature_end_time,
            )
            return format_sql(sql_str)
        else:
            msg = f"Unsupported aggregation level: {aggregation_level}"
            raise ValueError(msg)

    else:
        return format_sql(materialized_sql)


# By default Snowflake unloads numerical types as byte arrays because they have higher precision than available Parquet
# types. We could decode these in the OnlineStoreCopier, but since we will be downcasting them to the precision of
# Parquet types we might as well do it now.
#
# Snowflake docs for how casting interacts with Parquet types:
# https://docs.snowflake.com/en/user-guide/data-unload-considerations.html#explicitly-converting-numeric-columns-to-parquet-data-types
COPY_CASTS = {
    data_types.Int64Type(): "BIGINT",
    data_types.Float64Type(): "DOUBLE",
}


def get_materialization_query(
    feature_definition: FeatureDefinitionWrapper,
    source: str,
    # start is inclusive and end is exclusive
    feature_start_time: Optional[datetime] = None,
    feature_end_time: Optional[datetime] = None,
    # Whether we are building offline materialization query. If we are building an materialization query, in continuous
    # mode, we will have to hardcode the schedule interval to 1 day
    offline_materialization: bool = False,
):
    """Returns a SQL query for time-limited materialization.

    Does not include a terminating `;` or any COPY or INSERT statements."""

    if feature_definition.is_temporal_aggregate:
        source = MATERIALIZATION_TILE_TEMPLATE.render(
            source=source,
            join_keys=feature_definition.join_keys,
            aggregations=_get_feature_view_aggregations(feature_definition),
            slide_interval=feature_definition.aggregate_slide_interval,
            timestamp_key=feature_definition.time_key,
            continuous=feature_definition.is_continuous,
            offline_materialization=offline_materialization,
        )
    return format_sql(
        TIME_LIMIT_TEMPLATE.render(
            source=source,
            timestamp_key=feature_definition.time_key,
            start_time=feature_start_time,
            end_time=feature_end_time,
        )
    )


def get_delete_staged_files_sql(destination_stage: str, days: int):
    # Note the double quotes around destination_stage are required for the SQL to format properly
    script_sql = DELETE_STAGED_FILES_TEMPLATE.render(
        destination_stage=f"'{destination_stage}'",
        days=days,
    )
    sql = "\n".join(("EXECUTE IMMEDIATE", "$$", format_sql(script_sql), "$$;"))
    return format_sql(sql)


def get_delete_orphaned_schemas_sql(snowflake_database: str):
    script_sql = DELETE_ORPHANED_SCHEMAS_TEMPLATE.render(
        snowflake_database=snowflake_database,
    )
    sql = "\n".join(("EXECUTE IMMEDIATE", "$$", format_sql(script_sql), "$$;"))
    return format_sql(sql)


def get_materialization_copy_sql(
    feature_definition: FeatureDefinitionWrapper,
    # start is inclusive and end is exclusive
    time_limits: pendulum.Period,
    destination_stage: Optional[str],
    destination_table: Optional[str],
    # this is materialization task id; it's probably being used incorrectly because
    # different attempts can share the same task id.
    materialization_id: str,
    session: "snowflake.snowpark.Session" = None,
    # If provided, this query will be used instead of the feature_definition
    materialization_query: Optional[str] = None,
):
    """
    TODO(TEC-16752): Deprecate this once snowpark support is fully removed.
    Returns a SQL query for a COPY INTO an destination_stage for materialization.
    Additionally INSERTs into destination_table for manually-materialized FeatureViews.
    """

    materialization_context = BoundMaterializationContext._create_internal(
        time_limits.start, time_limits.end, feature_definition.fv_spec.batch_schedule
    )
    materialize_online = destination_stage is not None
    materialize_offline = destination_table is not None
    if materialization_query is None:
        source = pipeline_to_sql_string(
            pipeline=feature_definition.pipeline,
            data_sources=feature_definition.data_sources,
            transformations=feature_definition.transformations,
            materialization_context=materialization_context,
            session=session,
        )

        materialization_query = get_materialization_query(
            source=source,
            feature_definition=feature_definition,
            feature_start_time=time_limits.start,
            feature_end_time=time_limits.end,
            offline_materialization=materialize_offline,
        )
    common_context = {
        "source": materialization_query,
        "materialize_online": materialize_online,
        "materialize_offline": materialize_offline,
        "destination_stage": destination_stage,
        "materialization_schema": feature_definition.materialization_schema.to_proto(),
        "materialization_id": materialization_id,
        "cast_types": COPY_CASTS,
    }
    if materialize_offline:
        view_name = feature_definition.fv_spec.snowflake_view_name
        database, schema, view = view_name.split(".")
        script_sql = MATERIALIZED_FEATURE_VIEW_TEMPLATE.render(
            destination_table=destination_table,
            workspace=f"{database}.{schema}",
            destination_view=view_name,
            start_time=time_limits.start,
            end_time=time_limits.end,
            continuous=feature_definition.is_continuous,
            timestamp_field=feature_definition.time_key,
            is_aggregate=feature_definition.is_temporal_aggregate,
            slide_interval=feature_definition.aggregate_slide_interval
            if feature_definition.is_temporal_aggregate
            else duration_pb2.Duration(),
            **common_context,
        )
        sql = "\n".join(("EXECUTE IMMEDIATE", "$$", format_sql(script_sql), "$$;"))
    else:
        sql = ONLINE_STORE_COPIER_TEMPLATE.render(**common_context)
    return format_sql(sql)


def get_dataframe_for_data_source(
    session: "snowflake.snowpark.Session",
    data_source: specs.BatchSourceSpec,
    start_time: Optional[datetime] = None,
    end_time: Optional[datetime] = None,
) -> "snowflake.snowpark.DataFrame":
    assert isinstance(
        data_source, specs.SnowflakeSourceSpec
    ), f"Snowflake compute only supports Snowflake data sources. Got: {data_source}."

    if data_source.query:
        source = data_source.query
    else:
        source = f"{data_source.database}.{data_source.schema}.{data_source.table}"

    if (start_time is not None or end_time is not None) and data_source.timestamp_field is None:
        msg = "Filtering by start_time or end_time requires the timestamp_field parameter to be set on this data source"
        raise ValueError(msg)
    sql_str = TIME_LIMIT_TEMPLATE.render(
        source=source,
        timestamp_key=data_source.timestamp_field,
        start_time=start_time,
        end_time=end_time,
    )
    return session.sql(sql_str)


def _get_feature_view_aggregations(feature_definition: FeatureDefinitionWrapper) -> Dict[str, Set[str]]:
    aggregations = defaultdict(set)
    is_continuous = feature_definition.is_continuous
    # For continuous features, we only keep one aggregation function for first/last non distinct n for each input feature
    if is_continuous:
        max_first_distinct_n_aggs = defaultdict(int)
        max_last_distinct_n_aggs = defaultdict(int)

    for feature in feature_definition.fv_spec.aggregate_features:
        aggregate_function_lambda = AGGREGATION_PLANS[feature.function]
        if not aggregate_function_lambda:
            msg = f"Unsupported aggregation function {feature.function} in snowflake pipeline"
            raise TectonSnowflakeNotImplementedError(msg)
        if feature.function == afpb.AGGREGATION_FUNCTION_LAST_NON_DISTINCT_N:
            if is_continuous:
                max_last_distinct_n_aggs[feature.input_feature_name] = max(
                    max_last_distinct_n_aggs[feature.input_feature_name], feature.function_params.last_n.n
                )
                continue
            aggregate_function = aggregate_function_lambda(feature.function_params.last_n.n, is_continuous)
        elif feature.function == afpb.AGGREGATION_FUNCTION_FIRST_NON_DISTINCT_N:
            if is_continuous:
                max_first_distinct_n_aggs[feature.input_feature_name] = max(
                    max_first_distinct_n_aggs[feature.input_feature_name], feature.function_params.first_n.n
                )
                continue
            aggregate_function = aggregate_function_lambda(feature.function_params.first_n.n, is_continuous)
        else:
            aggregate_function = aggregate_function_lambda()
        aggregations[feature.input_feature_name].update(aggregate_function)

    if is_continuous:
        for feature_name, max_last_n in max_last_distinct_n_aggs.items():
            aggregations[feature_name].update(
                AGGREGATION_PLANS[afpb.AGGREGATION_FUNCTION_LAST_NON_DISTINCT_N](max_last_n, is_continuous)
            )
        for feature_name, max_first_n in max_first_distinct_n_aggs.items():
            aggregations[feature_name].update(
                AGGREGATION_PLANS[afpb.AGGREGATION_FUNCTION_FIRST_NON_DISTINCT_N](max_first_n, is_continuous)
            )

    # Need to order the functions for deterministic results.
    for key, value in aggregations.items():
        aggregations[key] = sorted(value)

    return aggregations



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/common/container_image__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n1tecton_proto/common/container_image__client.proto\x12\x13tecton_proto.common\"\x96\x01\n\x0e\x43ontainerImage\x12\'\n\x0frepository_name\x18\x01 \x01(\tR\x0erepositoryName\x12\x1b\n\timage_uri\x18\x02 \x01(\tR\x08imageUri\x12!\n\x0cimage_digest\x18\x03 \x01(\tR\x0bimageDigest\x12\x1b\n\timage_tag\x18\x04 \x01(\tR\x08imageTagB\x15\n\x11\x63om.tecton.commonP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.common.container_image__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\021com.tecton.commonP\001'
  _CONTAINERIMAGE._serialized_start=75
  _CONTAINERIMAGE._serialized_end=225
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/dataobs/config__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2
from tecton_proto.dataobs import expectation__client_pb2 as tecton__proto_dot_dataobs_dot_expectation____client__pb2
from tecton_proto.dataobs import metric__client_pb2 as tecton__proto_dot_dataobs_dot_metric____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n)tecton_proto/dataobs/config__client.proto\x12\x14tecton_proto.dataobs\x1a\x1egoogle/protobuf/duration.proto\x1a.tecton_proto/dataobs/expectation__client.proto\x1a)tecton_proto/dataobs/metric__client.proto\"\xa9\x03\n\x17\x44\x61taObservabilityConfig\x12\x1c\n\tworkspace\x18\x01 \x01(\tR\tworkspace\x12*\n\x11\x66\x65\x61ture_view_name\x18\x02 \x01(\tR\x0f\x66\x65\x61tureViewName\x12U\n\'feature_expectation_validation_schedule\x18\x03 \x01(\tR$featureExpectationValidationSchedule\x12\x36\n\x07metrics\x18\x04 \x03(\x0b\x32\x1c.tecton_proto.dataobs.MetricR\x07metrics\x12[\n\x14\x66\x65\x61ture_expectations\x18\x05 \x03(\x0b\x32(.tecton_proto.dataobs.FeatureExpectationR\x13\x66\x65\x61tureExpectations\x12X\n\x13metric_expectations\x18\x06 \x03(\x0b\x32\'.tecton_proto.dataobs.MetricExpectationR\x12metricExpectations\"\xb2\x01\n&DataObservabilityMaterializationConfig\x12\x18\n\x07\x65nabled\x18\x01 \x01(\x08R\x07\x65nabled\x12\x42\n\x0fmetric_interval\x18\x02 \x01(\x0b\x32\x19.google.protobuf.DurationR\x0emetricInterval\x12*\n\x11metric_table_name\x18\x03 \x01(\tR\x0fmetricTableNameB\x16\n\x12\x63om.tecton.dataobsP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.dataobs.config__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\022com.tecton.dataobsP\001'
  _DATAOBSERVABILITYCONFIG._serialized_start=191
  _DATAOBSERVABILITYCONFIG._serialized_end=616
  _DATAOBSERVABILITYMATERIALIZATIONCONFIG._serialized_start=619
  _DATAOBSERVABILITYMATERIALIZATIONCONFIG._serialized_end=797
# @@protoc_insertion_point(module_scope)



from typing import Any
from typing import Dict
from typing import List
from typing import Union

import pandas

from tecton_core import pipeline_common
from tecton_core import specs
from tecton_core.errors import UDF_ERROR
from tecton_core.id_helper import IdHelper
from tecton_proto.args.pipeline__client_pb2 import Pipeline
from tecton_proto.args.pipeline__client_pb2 import PipelineNode
from tecton_proto.args.pipeline__client_pb2 import TransformationNode
from tecton_proto.args.transformation__client_pb2 import TransformationMode


class _ODFVPipelineBuilder:
    def __init__(
        self,
        name: str,
        pipeline: Pipeline,
        transformations: List[specs.TransformationSpec],
    ) -> None:
        self._pipeline = pipeline
        self._name = name
        self._id_to_transformation = {t.id: t for t in transformations}
        self.mode = (
            "python"
            if self._id_to_transformation[
                IdHelper.to_string(self._pipeline.root.transformation_node.transformation_id)
            ].transformation_mode
            == TransformationMode.TRANSFORMATION_MODE_PYTHON
            else "pandas"
        )

    def _apply_transformation_function(
        self,
        transformation_node: TransformationNode,
        args: List[Union[pipeline_common.CONSTANT_TYPE, Dict[str, Any], pandas.DataFrame, pandas.Series]],
        kwargs: Dict[str, Any],
    ) -> Union[Dict[str, Any], pandas.DataFrame]:
        """For the given transformation node, returns the corresponding DataFrame transformation."""
        transformation = self._id_to_transformation[IdHelper.to_string(transformation_node.transformation_id)]
        user_function = transformation.user_function
        if transformation.transformation_mode == TransformationMode.TRANSFORMATION_MODE_PANDAS:
            try:
                res = user_function(*args, **kwargs)
            except Exception as e:
                raise UDF_ERROR(e, feature_definition_name=transformation.metadata.name) from e
            pipeline_common.check_transformation_type(transformation.name, res, "pandas", self._possible_modes())
            return res
        elif transformation.transformation_mode == TransformationMode.TRANSFORMATION_MODE_PYTHON:
            try:
                res = user_function(*args, **kwargs)
            except Exception as e:
                raise UDF_ERROR(e, feature_definition_name=transformation.metadata.name) from e
            # Only restrict types on the root node of python-mode transforms
            if transformation_node == self._pipeline.root:
                pipeline_common.check_transformation_type(transformation.name, res, "python", self._possible_modes())
            return res
        else:
            msg = f"Unknown transformation mode: {transformation.transformation_mode}"
            raise KeyError(msg)

    def _transformation_node_to_online_dataframe(
        self, transformation_node: TransformationNode
    ) -> Union[Dict[str, Any], pandas.DataFrame]:
        """Recursively translates inputs to values and then passes them to the transformation."""
        args: List[Union[str, int, float, bool, None, Dict[str, Any], pandas.DataFrame, pandas.Series]] = []
        kwargs = {}
        for transformation_input in transformation_node.inputs:
            node_value = self._udf_node_to_value(transformation_input.node)
            if transformation_input.HasField("arg_index"):
                assert len(args) == transformation_input.arg_index
                args.append(node_value)
            elif transformation_input.HasField("arg_name"):
                kwargs[transformation_input.arg_name] = node_value
            else:
                msg = f"Unknown argument type for Input node: {transformation_input}"
                raise KeyError(msg)

        return self._apply_transformation_function(transformation_node, args, kwargs)

    # evaluate a node in the Pipeline
    def _udf_node_to_value(
        self, pipeline_node: PipelineNode
    ) -> Union[pipeline_common.CONSTANT_TYPE, Dict[str, Any], pandas.DataFrame, pandas.Series]:
        if pipeline_node.HasField("constant_node"):
            return pipeline_common.constant_node_to_value(pipeline_node.constant_node)
        elif pipeline_node.HasField("feature_view_node"):
            if pipeline_node.feature_view_node.input_name not in self._passed_in_inputs:
                msg = f"Expected to find input {pipeline_node.feature_view_node.input_name} in provided ODFV pipeline inputs"
                raise ValueError(msg)

            return self._passed_in_inputs[pipeline_node.feature_view_node.input_name]
        elif pipeline_node.HasField("request_data_source_node"):
            if pipeline_node.request_data_source_node.input_name not in self._passed_in_inputs:
                msg = f"Expected to find input {pipeline_node.request_data_source_node.input_name} in provided ODFV pipeline inputs"
                raise ValueError(msg)

            return self._passed_in_inputs[pipeline_node.request_data_source_node.input_name]
        elif pipeline_node.HasField("transformation_node"):
            return self._transformation_node_to_online_dataframe(pipeline_node.transformation_node)
        elif pipeline_node.HasField("materialization_context_node"):
            msg = "MaterializationContext is unsupported for pandas pipelines"
            raise ValueError(msg)
        else:
            msg = "This is not yet implemented"
            raise NotImplementedError(msg)

    def _possible_modes(self):
        # note that pipeline is included since this is meant to be a user hint, and it's
        # theoretically possible a pipeline wound up deeper than expected
        return ["pandas", "pipeline", "python"]

    def execute_with_inputs(
        self, inputs: Union[Dict[str, pandas.DataFrame], Dict[str, Any]]
    ) -> Union[pipeline_common.CONSTANT_TYPE, Dict[str, Any], pandas.DataFrame, pandas.Series]:
        self._passed_in_inputs = inputs
        return self._udf_node_to_value(self._pipeline.root)


def build_odfv_execution_pipeline(
    pipeline: Pipeline,
    transformations: List[specs.TransformationSpec],
    name: str,
) -> _ODFVPipelineBuilder:
    return _ODFVPipelineBuilder(
        name=name,
        pipeline=pipeline,
        transformations=transformations,
    )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/databricks_api/error__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n/tecton_proto/databricks_api/error__client.proto\x12\x1btecton_proto.databricks_api\"^\n\rErrorResponse\x12\x1d\n\nerror_code\x18\x01 \x01(\tR\terrorCode\x12\x18\n\x07message\x18\x02 \x01(\tR\x07message\x12\x14\n\x05\x65rror\x18\x03 \x01(\tR\x05\x65rrorB\x1d\n\x19\x63om.tecton.databricks_apiP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.databricks_api.error__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\031com.tecton.databricks_apiP\001'
  _ERRORRESPONSE._serialized_start=80
  _ERRORRESPONSE._serialized_end=174
# @@protoc_insertion_point(module_scope)



import io
import os
import sys
import tarfile
from pathlib import Path

import click
import requests

from tecton import tecton_context
from tecton._internals import metadata_service
from tecton.cli import cli_utils
from tecton.cli import printer
from tecton.cli import repo_config
from tecton.cli.command import TectonCommand
from tecton_core import http
from tecton_core import repo_file_handler
from tecton_proto.metadataservice import metadata_service__client_pb2 as metadata_service_pb2


def init_feature_repo() -> None:
    if Path().resolve() == Path.home():
        printer.safe_print("You cannot set feature repository root to the home directory", file=sys.stderr)
        sys.exit(1)

    # If .tecton exists in a parent or child directory, error out.
    repo_root = repo_file_handler._maybe_get_repo_root()
    if repo_root not in [Path().resolve(), None]:
        printer.safe_print(".tecton already exists in a parent directory:", repo_root)
        sys.exit(1)

    child_dir_matches = list(Path().rglob("*/.tecton"))
    if len(child_dir_matches) > 0:
        dirs_str = "\n\t".join((str(c.parent.resolve()) for c in child_dir_matches))
        printer.safe_print(f".tecton already exists in child directories:\n\t{dirs_str}")
        sys.exit(1)

    dot_tecton = Path(".tecton")
    if not dot_tecton.exists():
        dot_tecton.touch()

        default_repo_config = Path("repo.yaml")
        if default_repo_config.exists():
            printer.safe_print(f"Found a repo config at {default_repo_config}. Skipping creating a new repo config.")
        else:
            repo_config.create_starter_repo_config(config_path=default_repo_config)

        msgs = [
            f"Local feature repository root set to {Path().resolve()}",
            "",
            "💡 We recommend tracking these files in git:",
            f"     {dot_tecton.resolve()}",
            f"     {default_repo_config.resolve()}",
            "",
            "💡 Run `tecton apply` to apply the feature repository to the Tecton cluster.",
        ]
        printer.safe_print("\n".join(msgs), file=sys.stderr)
    else:
        printer.safe_print("Feature repository is already set to", Path().resolve(), file=sys.stderr)


@click.command(uses_workspace=True, cls=TectonCommand)
@click.argument("commit_id", required=False)
@click.option("--yes", "-y", is_flag=True, default=False, help="Skip all confirmation prompts.")
def restore(commit_id, yes):
    """Restore feature repo state to a previous version.

    The commit to restore can either be passed as COMMIT_ID, or the latest commit for the currently selected workspace will be used.

    Also restores the repo config used for the apply to the default repo config path at REPO_ROOT/repo.yaml.
    """
    # Get the repo download URL from the metadata service.
    request = metadata_service_pb2.GetRestoreInfoRequest(workspace=tecton_context.get_current_workspace())
    if commit_id:
        request.commit_id = commit_id
    response = metadata_service.instance().GetRestoreInfo(request)

    # Download the repo.
    url = response.signed_url_for_repo_download
    commit_id = response.commit_id
    sdk_version = response.sdk_version
    # TODO: always print this message once enough customers are on new sdk versions
    sdk_version_msg = f"applied by SDK version {sdk_version}" if sdk_version else ""
    printer.safe_print(f"Restoring from commit {commit_id} {sdk_version_msg}")
    try:
        tar_response = http.session().get(url)
        tar_response.raise_for_status()
    except requests.RequestException as e:
        raise SystemExit(e)

    # Find the repo root or initialize a default repo if not in a repo.
    root = repo_file_handler._maybe_get_repo_root()
    if not root:
        init_feature_repo()
        root = Path().resolve()
    repo_file_handler.ensure_prepare_repo()

    # Get user confirmation for files that may be modified (all eligible python files or the default repo config path).
    existing_files = repo_file_handler.repo_files()

    default_repo_path = repo_config.get_default_repo_config_path()
    if default_repo_path.exists():
        existing_files.append(default_repo_path)

    if len(existing_files) > 0:
        if not yes:
            for f in existing_files:
                printer.safe_print(f)
            cli_utils.confirm_or_exit("This operation may delete or modify the above files. Ok?")
        for f in existing_files:
            os.remove(f)

    # Extract the feature repo.
    with tarfile.open(fileobj=io.BytesIO(tar_response.content), mode="r|gz") as tar:
        for entry in tar:
            if os.path.isabs(entry.name) or ".." in entry.name:
                msg = "Illegal tar archive entry"
                raise ValueError(msg)
            elif os.path.exists(root / Path(entry.name)):
                msg = f"tecton restore would overwrite an unexpected file: {entry.name}"
                raise ValueError(msg)
            tar.extract(entry, path=root)
    printer.safe_print("Success")


@click.command(requires_auth=False, cls=TectonCommand)
def init() -> None:
    """Initialize feature repo."""
    init_feature_repo()



"""Data models related to the Repo Config (i.e. the repo.yaml file).

See PRD: https://www.notion.so/tecton/PRD-Repo-Config-725bd10a7ce6422eaedaf8786869ea35
"""

import logging
from typing import Optional
from typing import Union

from tecton._internals.tecton_pydantic import StrictModel
from tecton._internals.tecton_pydantic import pydantic_v1
from tecton.framework import configs


logger = logging.getLogger(__name__)


def _number_to_string(value):
    if isinstance(value, (int, float)):
        return str(value)
    return value


class BatchFeatureViewDefaults(StrictModel):
    tecton_materialization_runtime: Optional[str] = None
    # TODO(jake): The online store default is currently set at data proto creation time based on backend state. This
    # is a very fragile, inconsistent approach that we should reconsider.
    online_store: Optional[configs.OnlineStoreTypes] = pydantic_v1.Field(default=None, discriminator="kind")
    # TODO(jake): Pydantic has several "union modes" - we should strongly prefer discriminated (i.e. using
    #  discriminator) unions because they have better error messages and make that the default or enforce in tests.
    offline_store: Union[configs.OfflineStoreConfig, configs.DeltaConfig, configs.ParquetConfig] = pydantic_v1.Field(
        default_factory=configs.OfflineStoreConfig, discriminator="kind"
    )
    batch_compute: configs.ComputeConfigTypes = pydantic_v1.Field(
        default_factory=configs._DefaultClusterConfig, discriminator="kind"
    )
    environment: Optional[str] = None

    @property
    def offline_store_config(self) -> configs.OfflineStoreConfig:
        if isinstance(self.offline_store, (configs.DeltaConfig, configs.ParquetConfig)):
            return configs.OfflineStoreConfig(staging_table_format=self.offline_store)
        else:
            return self.offline_store


class StreamFeatureViewDefaults(BatchFeatureViewDefaults):
    stream_compute: configs.ComputeConfigTypes = pydantic_v1.Field(
        default_factory=configs._DefaultClusterConfig, discriminator="kind"
    )


class FeatureTableDefaults(BatchFeatureViewDefaults):
    pass  # Currently the same as BatchFeatureViewDefaults.


class FeatureServiceDefaults(StrictModel):
    on_demand_environment: Optional[str] = None


class TectonObjectDefaults(StrictModel):
    batch_feature_view: Optional[BatchFeatureViewDefaults] = None
    stream_feature_view: Optional[StreamFeatureViewDefaults] = None
    feature_table: Optional[FeatureTableDefaults] = None
    feature_service: Optional[FeatureServiceDefaults] = None


class RepoConfig(StrictModel):
    """The data model for the repo config (i.e. the repo.yaml) file."""

    defaults: Optional[TectonObjectDefaults] = None


# Singleton Repo Config.
_repo_config: Optional[RepoConfig] = None


def set_repo_config(repo_config: RepoConfig) -> None:
    """Set the singleton instance of the repo config."""
    global _repo_config
    if _repo_config is not None:
        logger.warning("Overwriting Tecton repo config that was already set.")
    _repo_config = repo_config


def get_repo_config() -> Optional[RepoConfig]:
    """Get the singleton instance of the repo config. None if it has not been set.

    The repo config is expected to be None in non-plan/apply environments (e.g. notebooks) or if no config is found
    during plan/apply.
    """
    return _repo_config


def get_feature_service_defaults() -> FeatureServiceDefaults:
    """Get the user-specified Feature Service defaults or a default instance if the user did not specify defaults."""
    if _repo_config is None or _repo_config.defaults is None or _repo_config.defaults.feature_service is None:
        return FeatureServiceDefaults()

    return _repo_config.defaults.feature_service


def get_batch_feature_view_defaults() -> BatchFeatureViewDefaults:
    """Get the user-specified Batch FV defaults or a default instance if the user did not specify defaults."""
    if _repo_config is None or _repo_config.defaults is None or _repo_config.defaults.batch_feature_view is None:
        return BatchFeatureViewDefaults()

    return _repo_config.defaults.batch_feature_view


def get_stream_feature_view_defaults() -> StreamFeatureViewDefaults:
    """Get the user-specified Stream FV defaults or a default instance if the user did not specify defaults."""
    if _repo_config is None or _repo_config.defaults is None or _repo_config.defaults.stream_feature_view is None:
        return StreamFeatureViewDefaults()

    return _repo_config.defaults.stream_feature_view


def get_feature_table_defaults() -> FeatureTableDefaults:
    """Get the user-specified Feature Table defaults or a default instance if the user did not specify defaults."""
    if _repo_config is None or _repo_config.defaults is None or _repo_config.defaults.feature_table is None:
        return FeatureTableDefaults()

    return _repo_config.defaults.feature_table



"""
Functions for serializing Python functions.

The only known areas where we will fail to serialize something usable are:

user imports like import foo.bar.baz instead of from foo.bar import baz
are not detected properly by the underlying library we use for looking up
dependencies. import json or other top-level modeuls still works just fine,
its just when you include multiple path components that we don't detect the
dependency properly

references to instantiated non-builtin types with a repr that can't be exec'd

users referring to a renamed version of an imported module has some wonky
behavior, for example `import json` then `foo=json` then referring to
`foo` in the func to be serialized.
"""

import inspect
import logging
from collections import defaultdict
from types import FunctionType
from types import ModuleType
from typing import Callable

from tecton.vendor.dill.dill.detect import freevars
from tecton.vendor.dill.dill.detect import globalvars
from tecton_core import conf
from tecton_core import repo_file_handler
from tecton_core.errors import TectonValidationError
from tecton_proto.args import user_defined_function__client_pb2 as user_defined_function_pb2


WINDOW_UNBOUNDED_PRECEDING = "unbounded_preceding"

logger = logging.getLogger(__name__)


def to_proto(transform: Callable) -> user_defined_function_pb2.UserDefinedFunction:
    if hasattr(transform, "_code"):
        code = transform._code
    else:
        code = _getsource(transform)

    _validate_code(code)

    return user_defined_function_pb2.UserDefinedFunction(name=transform.__name__, body=code)


BANNED_INDIRECT_IMPORTS = ["materialization_context", "tecton_sliding_window_udf", "WINDOW_UNBOUNDED_PRECEDING"]


def _validate_code(code):
    for banned_import in BANNED_INDIRECT_IMPORTS:
        if f"tecton.{banned_import}" in code:
            msg = f"Cannot serialize `tecton.{banned_import}`. Please use an import like `from tecton import {banned_import}`. Note that even comments may trigger this if your code contains `tecton.{banned_import}`"
            raise Exception(msg)


# If these are in the path to a module, we don't expect the function definitions to be inside a feature repo, but we still want to inline them
# integration_tests/ is added as a hack because our integration tests aren't run in a feature repo. But unless a customer is importing from a python package that has a directory named integration_tests, it
# shouldn't cause any wrong behavior.
# IMPORTANT: Include backslash paths for windows.
TECTON_TRANSFORMATION_WHITELIST = ["tecton/compat", "tecton\compat", "integration_tests/"]


# returns true if the module is defined by tecton, rather than by the user
def is_tecton_defined(module):
    return any(whitelist_str in module for whitelist_str in TECTON_TRANSFORMATION_WHITELIST)


def should_serialize_function(func):
    """Determines if the function should be serialized or not.

    During notebook development, not all functions need to be serialized since they never need to be sent over the wire.

    See https://www.notion.so/tecton/Micro-RFC-Function-Serialization-for-Notebook-Driven-Development-9385338c4fd144758fe3dc9587351142
    """
    # TECTON_FORCE_FUNCTION_SERIALIZATION is used for internal testing and also as a potential escape hatch for
    # notebook dev users.
    forced_behavior = conf.get_or_none("TECTON_FORCE_FUNCTION_SERIALIZATION")
    if forced_behavior is not None:
        if forced_behavior.lower() == "true":
            return True
        elif forced_behavior.lower() == "false":
            return False
        else:
            msg = "TECTON_FORCE_FUNCTION_SERIALIZATION should be 'true', 'false', or unset."
            raise ValueError(msg)

    func_module = inspect.getmodule(func)
    if func_module is None:
        # Used by NDD with EMR. In EMR pyspark notebook, 'inspect.getmodule' returns None for functions defined in main scope.
        return False

    if func_module.__name__ == "__main__":
        # This function was defined in the main module (i.e. probably in a notebook or repl). Do not attempt to
        # serialize it.
        return False

    func_file = inspect.getfile(func_module)

    if is_tecton_defined(func_file):
        # This is a Tecton defined transformation. It should always be serialized.
        return True

    if repo_file_handler.is_file_in_a_tecton_repo(func_file):
        # If a function is in a Tecton repo (i.e. an ancestor directory has a .tecton file), then always attemp to
        # serialize it. In notebook development, this will occur if an FCO is imported into the notebook from a repo.
        repo_file_handler.ensure_prepare_repo(func_file)
        if func_file not in repo_file_handler.repo_files_set():
            msg = f'Error during function serialization. The file "{func_file}" is in a Tecton repo but not in the cached repo data. Importing Tecton objects from multiple repos is not supported.'
            raise TectonValidationError(msg)
        return True
    else:
        # This function is defined outside of the main module, but not in a Tecton repo. Do not attempt to serialize it.
        return False


# UNDONE:
# this breaks for the following case:
# def f():
#    a = module.CONSTANT
#
# for some reason, the module/constant don't show up in globalvars.
# this case has not occurred in practice
def _getsource(func):
    imports = defaultdict(set)
    modules = set()
    code_lines = []
    seen_args = {}

    # file the func was defined in
    func_file = inspect.getfile(inspect.getmodule(func))

    # Check if the file is part of the tecton codebase, not the feature repo.
    # Only perform any feature repo initialization for functions in the feature repo.
    # We short-circuit any repo_file_handler functions that depend on initialization if we skip initialization here.
    if not is_tecton_defined(func_file):
        repo_file_handler.ensure_prepare_repo(func_file)

    def process_functiontype(name, obj, imports, modules, code_lines, seen_args):
        # if this is user-defined or otherwise unavailable to import at deserialization time
        # including anything without a module, with module of __main__, defined inside the feature repo
        # or anything with a qualified name containing characters invalid for an import path, like foo.<locals>.bar
        module = inspect.getfile(inspect.getmodule(obj))
        if (
            obj.__module__ in ("__main__", None)
            or "<" in obj.__qualname__
            or is_tecton_defined(module)
            or module in repo_file_handler.repo_files_set()
        ):
            objs = globalvars(obj, recurse=False)
            objs.update(freevars(obj))
            default_objs = {}
            for param in inspect.signature(obj).parameters.values():
                if param.default != inspect.Parameter.empty:
                    default_objs[param.name] = param.default
                    # these defaults shadow over obj
                    objs.pop(param.name, None)
            # need to sort the keys since globalvars ordering is non-deterministic
            for dependency, dep_obj in sorted(objs.items()):
                recurse(dependency, dep_obj, imports, modules, code_lines, seen_args, write_codelines=True)
            for dependency, dep_obj in sorted(default_objs.items()):
                # we dont re-write defaults at top-level since the `def` lines should have the declarations
                recurse(dependency, dep_obj, imports, modules, code_lines, seen_args, write_codelines=False)
            fdef = inspect.getsource(obj)
            fdef = fdef[fdef.find("def ") :]
            code_lines.append(fdef)
        else:
            imports[obj.__module__].add(obj.__name__)

    def recurse(name, obj, imports, modules, code_lines, seen_args, write_codelines):
        def _add_codeline(line):
            if write_codelines:
                code_lines.append(line)

        # prevent processing same dependency object multiple times, even if
        # multiple dependent objects exist in the tree from the original
        # func
        seen_key = str(name) + str(obj)
        if seen_args.get(seen_key) is True:
            return
        seen_args[seen_key] = True

        has_spark = True
        try:
            import pyspark  # noqa: F401
        except ImportError:
            has_spark = False

        is_spark_struct = False
        is_unbound_materialization_context = False
        is_tecton_sliding_window_udf = False
        if has_spark:
            from pyspark.sql import types as spark_types

            is_spark_struct = isinstance(obj, spark_types.StructType)
            from tecton_core.materialization_context import UnboundMaterializationContext

            is_unbound_materialization_context = isinstance(obj, UnboundMaterializationContext)

            from tecton_spark.udfs import tecton_sliding_window_udf

            is_tecton_sliding_window_udf = obj is tecton_sliding_window_udf

        # Confusingly classes are subtypes of 'type'; non-classes are not
        if isinstance(obj, type):
            if obj.__module__ == "__main__":
                msg = f"Cannot serialize class {obj.__name__} from module __main__"
                raise Exception(msg)
            imports[obj.__module__].add(obj.__name__)

        elif isinstance(obj, FunctionType):
            process_functiontype(name, obj, imports, modules, code_lines, seen_args)
        elif isinstance(obj, ModuleType):
            module_file = inspect.getfile(obj)
            # skip this check for functions written by Tecton
            if not is_tecton_defined(func_file) and module_file in repo_file_handler.repo_files_set():
                msg = f"Cannot serialize usage of {obj.__name__}. Instead, directly import objects from the module, e.g. `from {obj.__name__} import obj`"
                raise Exception(msg)
            if f"{obj.__package__}.{name}" == obj.__name__:
                imports[obj.__package__].add(name)
            else:
                modules.add(obj.__name__)
        elif is_spark_struct:
            _add_codeline(f"{name} = StructType.fromJson(json.loads('{obj.json()}'))")
            modules.add("json")
            imports["pyspark.sql.types"].add("StructType")
        elif is_unbound_materialization_context:
            imports["tecton_core.materialization_context"].add("materialization_context")
        elif is_tecton_sliding_window_udf:
            imports["tecton_spark.udfs"].add("tecton_sliding_window_udf")
        elif obj == WINDOW_UNBOUNDED_PRECEDING:
            imports["tecton_spark.time_utils"].add("WINDOW_UNBOUNDED_PRECEDING")
        else:
            try:
                repr_str = f"{name}={repr(obj)}"
                exec(repr_str)
                _add_codeline(repr_str)
            except Exception:
                msg = f"Cannot evaluate object {obj} of type '{type(obj)}' for serialization"
                raise Exception(msg)

    recurse(func.__name__, func, imports, modules, code_lines, seen_args, write_codelines=True)

    for module in sorted(imports):
        import_line = f"from {module} import "
        import_line += ", ".join(sorted(imports[module]))
        code_lines.insert(0, import_line)

    for module in sorted(modules):
        code_lines.insert(0, f"import {module}")

    return "\n".join(code_lines)


def inlined(func):
    # this a no-op that can be deleted later
    return func



# This package is built to improve the error message of typeguard 2.0. We can deprecate this module once we upgrade the
# `typeguard` to 4.0+. Details are avaiable in https://tecton.atlassian.net/browse/TEC-19929.

from functools import wraps

from typeguard import typechecked

from tecton.framework import feature


_BATCH_FEATURE_VIEW = "batch_feature_view"
_STREAM_FEATURE_VIEW = "stream_feature_view"
_REALTIME_FEATURE_VIEW = "realtime_feature_view"


def batch_feature_view_typechecked(f):
    return _internal_typechecked(f, _BATCH_FEATURE_VIEW)


def stream_feature_view_typechecked(f):
    return _internal_typechecked(f, _STREAM_FEATURE_VIEW)


def realtime_feature_view_typechecked(f):
    return _internal_typechecked(f, _REALTIME_FEATURE_VIEW)


def _internal_typechecked(f, fco_type):
    decorated_func = typechecked(f)

    @wraps(f)
    def wrapper(*args, **kwargs):
        if "features" in kwargs:
            for i, item in enumerate(kwargs["features"]):
                if fco_type == _BATCH_FEATURE_VIEW and not isinstance(item, feature.Feature):
                    msg = f"`features` expects a list of `Aggregate` or a list of `Union[Attribute, Embedding]`, but features[{i}] is '${item.__class__}'"
                    raise TypeError(msg)
                if fco_type == _STREAM_FEATURE_VIEW and not isinstance(item, feature.Feature):
                    msg = f"`features` expects a list of `Aggregate` or a list of `Attribute`, but features[{i}] is '${item.__class__}'"
                    raise TypeError(msg)
                if fco_type == _REALTIME_FEATURE_VIEW and not isinstance(item, feature.Attribute):
                    msg = f"`features` expects a list of `Attribute`, but features[{i}] is '${item.__class__}'"
                    raise TypeError(msg)
        return decorated_func(*args, **kwargs)

    return wrapper



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2011
# Brett Alistair Kromkamp - brettkromkamp@gmail.com
# Copyright (C) 2012-2017
# Xiaming Chen - chenxm35@gmail.com
# and other contributors.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This file contains modifications by Tecton, Inc.
"""
This is a public location to maintain contributed
utilities to extend the basic Tree class.

Deprecated! We prefer a unified processing of Tree object.
"""
from .misc import deprecated


@deprecated(alias='tree.to_graphviz()')
def export_to_dot(tree, filename=None, shape='circle', graph='digraph'):
    """Exports the tree in the dot format of the graphviz software"""
    tree.to_graphviz(filename=filename, shape=shape, graph=graph)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/data/materialization_roles_allowlists__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n@tecton_proto/data/materialization_roles_allowlists__client.proto\x12\x11tecton_proto.data\"\x96\x01\n\x1eMaterializationRolesAllowlists\x12t\n\x18workspace_configurations\x18\x01 \x03(\x0b\x32\x39.tecton_proto.data.WorkspaceMaterializationRolesAllowlistR\x17workspaceConfigurations\"\x93\x01\n&WorkspaceMaterializationRolesAllowlist\x12%\n\x0eworkspace_name\x18\x01 \x01(\tR\rworkspaceName\x12\x42\n\x1d\x61llowed_materialization_roles\x18\x02 \x03(\tR\x1b\x61llowedMaterializationRolesB\x13\n\x0f\x63om.tecton.dataP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.data.materialization_roles_allowlists__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.dataP\001'
  _MATERIALIZATIONROLESALLOWLISTS._serialized_start=88
  _MATERIALIZATIONROLESALLOWLISTS._serialized_end=238
  _WORKSPACEMATERIALIZATIONROLESALLOWLIST._serialized_start=241
  _WORKSPACEMATERIALIZATIONROLESALLOWLIST._serialized_end=388
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/metadataservice/http_over_grpc__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.auth import service__client_pb2 as tecton__proto_dot_auth_dot_service____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n9tecton_proto/metadataservice/http_over_grpc__client.proto\x12\x1ctecton_proto.metadataservice\x1a\'tecton_proto/auth/service__client.proto\"\x93\x01\n\x0bHttpRequest\x12\x16\n\x06method\x18\x01 \x01(\tR\x06method\x12\x10\n\x03url\x18\x02 \x01(\tR\x03url\x12\x12\n\x04\x62ody\x18\x03 \x01(\x0cR\x04\x62ody\x12 \n\x0b\x63ontentType\x18\x04 \x01(\tR\x0b\x63ontentType\x12$\n\rcontentLength\x18\x05 \x01(\x03R\rcontentLength\"u\n\x13HttpResponseHeaders\x12\x16\n\x06status\x18\x01 \x01(\x05R\x06status\x12 \n\x0b\x63ontentType\x18\x02 \x01(\tR\x0b\x63ontentType\x12$\n\rcontentLength\x18\x03 \x01(\x03R\rcontentLength\"\x93\x01\n\x11HttpResponsePiece\x12M\n\x07headers\x18\x01 \x01(\x0b\x32\x31.tecton_proto.metadataservice.HttpResponseHeadersH\x00R\x07headers\x12\x1f\n\nbody_piece\x18\x02 \x01(\x0cH\x00R\tbodyPieceB\x0e\n\x0cmessage_type2\xac\x01\n\x13HttpOverGrpcService\x12\x94\x01\n\x05Proxy\x12).tecton_proto.metadataservice.HttpRequest\x1a/.tecton_proto.metadataservice.HttpResponsePiece\"-\xa2\xbc\xe6\xc0\x05\'*%tecton/organization#read_organization0\x01\x42_\n\x1a\x63om.tecton.metadataserviceB\x0cHttpOverGrpcP\x01Z1github.com/tecton-ai/tecton_proto/metadataservice')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.metadataservice.http_over_grpc__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\032com.tecton.metadataserviceB\014HttpOverGrpcP\001Z1github.com/tecton-ai/tecton_proto/metadataservice'
  _HTTPOVERGRPCSERVICE.methods_by_name['Proxy']._options = None
  _HTTPOVERGRPCSERVICE.methods_by_name['Proxy']._serialized_options = b'\242\274\346\300\005\'*%tecton/organization#read_organization'
  _HTTPREQUEST._serialized_start=133
  _HTTPREQUEST._serialized_end=280
  _HTTPRESPONSEHEADERS._serialized_start=282
  _HTTPRESPONSEHEADERS._serialized_end=399
  _HTTPRESPONSEPIECE._serialized_start=402
  _HTTPRESPONSEPIECE._serialized_end=549
  _HTTPOVERGRPCSERVICE._serialized_start=552
  _HTTPOVERGRPCSERVICE._serialized_end=724
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/data/virtual_data_source__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.common import data_source_type__client_pb2 as tecton__proto_dot_common_dot_data__source__type____client__pb2
from tecton_proto.common import id__client_pb2 as tecton__proto_dot_common_dot_id____client__pb2
from tecton_proto.common import schema_container__client_pb2 as tecton__proto_dot_common_dot_schema__container____client__pb2
from tecton_proto.data import batch_data_source__client_pb2 as tecton__proto_dot_data_dot_batch__data__source____client__pb2
from tecton_proto.data import fco_metadata__client_pb2 as tecton__proto_dot_data_dot_fco__metadata____client__pb2
from tecton_proto.data import stream_data_source__client_pb2 as tecton__proto_dot_data_dot_stream__data__source____client__pb2
from tecton_proto.validation import validator__client_pb2 as tecton__proto_dot_validation_dot_validator____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n3tecton_proto/data/virtual_data_source__client.proto\x12\x11tecton_proto.data\x1a\x32tecton_proto/common/data_source_type__client.proto\x1a$tecton_proto/common/id__client.proto\x1a\x32tecton_proto/common/schema_container__client.proto\x1a\x31tecton_proto/data/batch_data_source__client.proto\x1a,tecton_proto/data/fco_metadata__client.proto\x1a\x32tecton_proto/data/stream_data_source__client.proto\x1a/tecton_proto/validation/validator__client.proto\"\xd2\x05\n\x11VirtualDataSource\x12L\n\x16virtual_data_source_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x13virtualDataSourceId\x12N\n\x11\x62\x61tch_data_source\x18\x04 \x01(\x0b\x32\".tecton_proto.data.BatchDataSourceR\x0f\x62\x61tchDataSource\x12Q\n\x12stream_data_source\x18\x05 \x01(\x0b\x32#.tecton_proto.data.StreamDataSourceR\x10streamDataSource\x12M\n\x10\x64\x61ta_source_type\x18\x06 \x01(\x0e\x32#.tecton_proto.common.DataSourceTypeR\x0e\x64\x61taSourceType\x12\x41\n\x0c\x66\x63o_metadata\x18\x07 \x01(\x0b\x32\x1e.tecton_proto.data.FcoMetadataR\x0b\x66\x63oMetadata\x12<\n\x06schema\x18\t \x01(\x0b\x32$.tecton_proto.common.SchemaContainerR\x06schema\x12\x61\n\x0fvalidation_args\x18\n \x01(\x0b\x32\x38.tecton_proto.validation.VirtualDataSourceValidationArgsR\x0evalidationArgs\x12K\n\x07options\x18\x0b \x03(\x0b\x32\x31.tecton_proto.data.VirtualDataSource.OptionsEntryR\x07options\x1a:\n\x0cOptionsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01J\x04\x08\x02\x10\x03J\x04\x08\x03\x10\x04J\x04\x08\x08\x10\tB\x13\n\x0f\x63om.tecton.dataP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.data.virtual_data_source__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.dataP\001'
  _VIRTUALDATASOURCE_OPTIONSENTRY._options = None
  _VIRTUALDATASOURCE_OPTIONSENTRY._serialized_options = b'8\001'
  _VIRTUALDATASOURCE._serialized_start=415
  _VIRTUALDATASOURCE._serialized_end=1137
  _VIRTUALDATASOURCE_OPTIONSENTRY._serialized_start=1061
  _VIRTUALDATASOURCE_OPTIONSENTRY._serialized_end=1119
# @@protoc_insertion_point(module_scope)



import logging
import signal
import time
from typing import Optional

from pyspark.sql import SparkSession
from pyspark.sql.functions import concat
from pyspark.sql.functions import from_json
from pyspark.sql.functions import lit
from pyspark.sql.functions import struct
from pyspark.sql.functions import to_json
from pyspark.sql.types import StringType
from pyspark.sql.types import StructType

from tecton_core.query_consts import anchor_time
from tecton_proto.materialization.params__client_pb2 import MaterializationTaskParams
from tecton_spark.materialization_plan import MATERIALIZED_RAW_DATA_END_TIME


logger = logging.getLogger(__name__)

WRITTEN_BY_BATCH = "written_by_batch"
BATCH_STATUS_ENTRY = "update_status_only"
COLUMNS = "columns"
CANARY_ID_COLUMN = "canary_id"


def batch_write_to_online_store(
    dataframe,
    materialization_task_params: MaterializationTaskParams,
    sink,
    fv_id: str,
    is_status: bool,
):
    start = time.time()
    canary_id = materialization_task_params.canary_id if materialization_task_params.HasField("canary_id") else None
    write_df = df_to_online_store_msg(dataframe, fv_id, is_batch=True, is_status=is_status, canary_id=canary_id)

    logger.info(f"Starting batch write to Tecton Online Store for the FV '{fv_id}'")
    write_df._jdf.foreachPartition(sink)

    latency = time.time() - start
    logger.info(f"Finished batch write for the FV '{fv_id}' ({latency}s)")


def df_to_online_store_msg(
    dataframe,
    feature_view_id_str: str,
    is_batch: bool,
    is_status: bool,
    is_compaction_job: bool = False,
    canary_id: Optional[str] = None,
):
    """Produces a dataframe to be written to the online store.

    The dataframe will have two columns: 'key' and 'value'. The 'key' column will look like:

    feature_view_id_str # for SFVs and SWAFVs
    or
    feature_view_id_str + "|" + _anchor_time # for BWAFVs
    or
    feature_view_id_str + "|" + _materialized_raw_data_end_time # for batch BFVs

    In order to construct these keys, the input dataframe must have a column named `_anchor_time` for BWAFVs or a column
    named `_materialized_raw_data_end_time` for BFVs. It does not need to have either for SFVs or SWAFVs.

    For BWAFVs, the 'value' column will look like:
        value = {WRITTEN_BY_BATCH: true, "_anchor_time": 196000, COLUMNS: {"num_users": 1, ...}} # `is_status` is False
        or
        value = {BATCH_STATUS_ENTRY: true, WRITTEN_BY_BATCH: true, "_anchor_time": 196000} # `is_status` is True
    For BFVs, the 'value' column will look like:
        value = {WRITTEN_BY_BATCH: true, "_materialized_raw_data_end_time": 196000, COLUMNS: {"num_users": 1, ...}} # `is_status` is False
        or
        value = {BATCH_STATUS_ENTRY: true, WRITTEN_BY_BATCH: true, "_materialized_raw_data_end_time": 196000} # `is_status` is True
    For SFVs and SWAFVs, the 'value' column will look like:
        value = {COLUMNS: {"num_users": 1, ...}} # `is_status` must be False when `is_batch` is False
    """
    # only batch sends data and status separately
    assert not is_status or is_batch or is_compaction_job

    # add additional columns if needed; wrap original columns except anchor time in COLUMNS struct
    payload_schema = dataframe.schema.fieldNames()
    key = lit(feature_view_id_str)
    if is_compaction_job:
        return dataframe.select(key.alias("key"), to_json(struct(payload_schema)).alias("value"))

    # Only used when `is_batch` is True.
    is_temporal_aggregate = False

    if anchor_time() in payload_schema:
        is_temporal_aggregate = True
        payload_schema.remove(anchor_time())
    elif is_batch:
        payload_schema.remove(MATERIALIZED_RAW_DATA_END_TIME)

    if is_status:
        dataframe = dataframe.withColumn(BATCH_STATUS_ENTRY, lit(True))
    else:
        dataframe = dataframe.withColumn(COLUMNS, struct(payload_schema))
    if is_batch:
        dataframe = dataframe.withColumn(WRITTEN_BY_BATCH, lit(True))
    if canary_id:
        dataframe = dataframe.withColumn(CANARY_ID_COLUMN, lit(canary_id))

    # Remove original columns (except anchor_time or materialized_raw_data_end_time) from
    # a top-level dataframe
    for col_name in payload_schema:
        dataframe = dataframe.drop(col_name)

    # wrap all columns in json object as `value`
    row_schema = dataframe.schema.fieldNames()
    if is_batch:
        if is_temporal_aggregate:
            key = concat(key, lit("|"), anchor_time())
        else:
            key = concat(key, lit("|"), MATERIALIZED_RAW_DATA_END_TIME)

    dataframe = dataframe.select(key.alias("key"), to_json(struct(row_schema)).alias("value"))
    if is_batch and is_status and not is_temporal_aggregate:
        logger.info(f"Writing status table update for BFV: {feature_view_id_str}")

    return dataframe


def set_up_online_store_sink(spark: SparkSession, materialization_task_params: MaterializationTaskParams):
    sink = spark._jvm.com.tecton.onlinestorewriter.SparkOnlineStoreSinkFactory.fromMaterializationTaskParams(
        materialization_task_params.SerializeToString()
    )
    signal.signal(signal.SIGINT, lambda signum, frame: sink.closeGlobalResources())
    signal.signal(signal.SIGTERM, lambda signum, frame: sink.closeGlobalResources())
    return sink


class MaterializationResourceContainer:
    def __init__(self, spark: SparkSession, materialization_task_params: MaterializationTaskParams):
        self.spark = spark
        self.materialization_task_params = materialization_task_params
        self.online_store_sink = None
        self.materialization_plan = {}

    def get_online_store_sink(self):
        if self.online_store_sink is None:
            self.online_store_sink = set_up_online_store_sink(self.spark, self.materialization_task_params)
        return self.online_store_sink

    def close(self):
        if self.online_store_sink is not None:
            self.online_store_sink.closeGlobalResources()


def _set_statsd_client_prefix(statsd_client, spark):
    # Calling the statsd client directly will always emit metrics from the driver.
    app_name = spark.conf.get("spark.app.name")
    statsd_client._prefix = f"spark.{app_name}.driver"


def get_statsd_client(spark: SparkSession):
    # TEMPORARY: statsd is a new library and requires restarting the internal
    # cluster, so we import it here and gate it behind a flag. This will allow
    # us to gradually roll it out to customers, so we don't have to restart all
    # clusters upon a single release.
    import statsd

    statsd_client = statsd.StatsClient("0.0.0.0", 3031)
    _set_statsd_client_prefix(statsd_client, spark)
    return statsd_client


def wait_for_metric_scrape():
    # Sleep for 1 minute before closing to ensure metrics are collected by Chronosphere (30 second
    # scrape interval).
    sleep_time = 60  # 1 minute
    logger.info(f"Waiting {sleep_time}s for metrics to be scraped.")
    time.sleep(sleep_time)


def has_prior_delta_commit(spark, materialization_params, idempotence_key, idempotence_value):
    assert materialization_params.HasField("offline_store_path"), "definition must have offline_store_path set"

    from delta.tables import DeltaTable
    from pyspark.sql.utils import AnalysisException

    try:
        delta_table = DeltaTable.forPath(spark, materialization_params.offline_store_path)
    except AnalysisException:
        # no prior commits if table doesn't exist yet
        return False

    metadata_schema = StructType().add(f"{idempotence_key}", StringType())
    commit_count = (
        delta_table.history()
        .select(from_json("userMetadata", metadata_schema).alias("metadataJson"))
        .filter(f"metadataJson.{idempotence_key} = '{idempotence_value}'")
        .count()
    )
    return commit_count != 0



from typing import Tuple

from tecton._internals import errors
from tecton.framework import base_tecton_object
from tecton_proto.args import transformation__client_pb2 as transformation__args_proto


SPARK_SQL_MODE = "spark_sql"
PYSPARK_MODE = "pyspark"
SNOWFLAKE_SQL_MODE = "snowflake_sql"
SNOWPARK_MODE = "snowpark"
PANDAS_MODE = "pandas"
PYTHON_MODE = "python"
BIGQUERY_SQL_MODE = "bigquery_sql"


def short_tecton_objects_repr(tecton_objects: Tuple[base_tecton_object.BaseTectonObject]) -> str:
    """Returns a shortened printable representation for a tuple of Tecton objects. Used for printing summaries."""
    short_strings = tuple(short_tecton_object_repr(obj) for obj in tecton_objects)
    return repr(short_strings)


def short_tecton_object_repr(tecton_object: base_tecton_object.BaseTectonObject) -> str:
    """Returns a shortened printable representation for a Tecton object. Used for printing summaries."""
    return f"{type(tecton_object).__name__}('{tecton_object.info.name}')"


def get_transformation_mode_enum(mode: str, name: str) -> transformation__args_proto.TransformationMode.ValueType:
    """Returns the TransformationMode type from string"""
    if mode == SPARK_SQL_MODE:
        return transformation__args_proto.TransformationMode.TRANSFORMATION_MODE_SPARK_SQL
    elif mode == PYSPARK_MODE:
        return transformation__args_proto.TransformationMode.TRANSFORMATION_MODE_PYSPARK
    elif mode == SNOWFLAKE_SQL_MODE:
        return transformation__args_proto.TransformationMode.TRANSFORMATION_MODE_SNOWFLAKE_SQL
    elif mode == SNOWPARK_MODE:
        return transformation__args_proto.TransformationMode.TRANSFORMATION_MODE_SNOWPARK
    elif mode == PANDAS_MODE:
        return transformation__args_proto.TransformationMode.TRANSFORMATION_MODE_PANDAS
    elif mode == PYTHON_MODE:
        return transformation__args_proto.TransformationMode.TRANSFORMATION_MODE_PYTHON
    elif mode == BIGQUERY_SQL_MODE:
        return transformation__args_proto.TransformationMode.TRANSFORMATION_MODE_BIGQUERY_SQL
    else:
        raise errors.InvalidTransformationMode(
            name,
            mode,
            [SPARK_SQL_MODE, PYSPARK_MODE, SNOWFLAKE_SQL_MODE, SNOWPARK_MODE, PANDAS_MODE, PYTHON_MODE],
        )



import functools
import os
import re
import sys
from difflib import unified_diff
from pathlib import Path
from typing import List
from typing import Optional
from typing import Tuple

import click
from attr import asdict
from colorama import Fore
from colorama import Style
from google.protobuf import empty_pb2

from tecton._internals import metadata_service
from tecton._internals.display import Displayable
from tecton.cli import printer
from tecton_core.errors import FailedPreconditionError
from tecton_core.errors import TectonAbortedError
from tecton_core.errors import TectonAlreadyExistsError
from tecton_core.errors import TectonAPIInaccessibleError
from tecton_core.errors import TectonAPIValidationError
from tecton_core.errors import TectonDeadlineExceededError
from tecton_core.errors import TectonInternalError
from tecton_core.errors import TectonNotFoundError
from tecton_core.errors import TectonNotImplementedError
from tecton_core.errors import TectonOperationCancelledError
from tecton_core.errors import TectonResourceExhaustedError


_CLIENT_VERSION_INFO_RESPONSE_HEADER = "x-tecton-client-version-info"
_CLIENT_VERSION_WARNING_RESPONSE_HEADER = "x-tecton-client-version-warning"


def human_fco_type(fco_type: str, plural=False) -> str:
    name_map = {
        "virtual_data_source": ("DataSource", "DataSources"),
        "batch_data_source": ("BatchDataSource", "BatchDataSources"),
        "stream_data_source": ("StreamDataSource", "StreamDataSources"),
        "push_source": ("PushSource", "PushSources"),
        "entity": ("Entity", "Entities"),
        "transformation": ("Transformation", "Transformations"),
        "feature_table": ("FeatureTable", "FeatureTables"),
        "feature_view": ("FeatureView", "FeatureViews"),
        "batch_feature_view": ("BatchFeatureView", "BatchFeatureViews"),
        "on_demand_feature_view": ("OnDemandFeatureView", "OnDemandFeatureViews"),
        "stream_feature_view": ("StreamFeatureView", "StreamFeatureViews"),
        "batch_window_aggregate_feature_view": ("BatchWindowAggregateFeatureView", "BatchWindowAggregateFeatureViews"),
        "stream_window_aggregate_feature_view": (
            "StreamWindowAggregateFeatureView",
            "StreamWindowAggregateFeatureViews",
        ),
        "feature_service": ("FeatureService", "FeatureServices"),
    }
    if plural:
        return name_map[fco_type][1]
    else:
        return name_map[fco_type][0]


def bold(x):
    return Style.BRIGHT + x + Style.NORMAL


def ask_user(message: str, options: List[str], default=None, let_fail=False) -> Optional[str]:
    options_idx = {o.lower(): i for i, o in enumerate(options)}

    while True:
        if len(options) > 1:
            printer.safe_print(message, "[" + "/".join(options) + "]", end="> ")
        else:
            printer.safe_print(message, end="> ")

        try:
            user_input = input().strip().lower()
        except EOFError:
            return None

        if user_input == "" and default:
            return default

        if user_input in options_idx:
            return options[options_idx[user_input]]
        else:
            # If there is only one input option, typing "!" will select it.
            if user_input == "!" and len(options) == 1:
                return options[0]
            elif let_fail:
                return None


def confirm_or_exit(message, expect=None):
    try:
        if expect:
            if ask_user(message, options=[expect], let_fail=True) is not None:
                return
            else:
                printer.safe_print("Aborting")
                sys.exit(1)
        else:
            if ask_user(message, options=["y", "N"], default="N") == "y":
                return
            else:
                printer.safe_print("Aborting")
                sys.exit(1)
    except KeyboardInterrupt:
        printer.safe_print("Aborting")
        sys.exit(1)


def color_line(x):
    if x.startswith("+"):
        return Fore.GREEN + x + Fore.RESET
    elif x.startswith("-"):
        return Fore.RED + x + Fore.RESET
    return x


def color_diff(lines):
    return map(color_line, lines)


def indent_line(lines, indent):
    return (" " * indent + x for x in lines)


# TODO: Reuse this in other places that does the same (engine.py)
def pprint_dict(kv, colwidth, indent=0):
    for k, v in kv.items():
        printer.safe_print(indent * " " + f"{k.ljust(colwidth)} {v}")


def pprint_attr_obj(key_map, obj, colwidth):
    o = asdict(obj)
    pprint_dict({key_map[key]: o[key] for key in o}, colwidth)


def code_diff(diff_item, indent):
    return re.split(
        "\n",
        "".join(
            indent_line(
                color_diff(
                    unified_diff(
                        diff_item.val_existing.splitlines(keepends=True),
                        diff_item.val_declared.splitlines(keepends=True),
                    )
                ),
                indent,
            )
        ),
        3,
    )[-1]


def print_version_msg(message, is_warning=False):
    if isinstance(message, list):
        message = message[-1] if len(message) > 0 else ""
    color = Fore.YELLOW
    if is_warning:
        message = "⚠️  " + message
    printer.safe_print(color + message + Fore.RESET, file=sys.stderr)


def display_principal(principal, default=""):
    principal_type = principal.WhichOneof("basic_info")

    if principal_type == "user":
        return principal.user.login_email
    if principal_type == "service_account":
        identifier = principal.service_account.name if principal.service_account.name else principal.service_account.id
        return identifier
    return default


def display_table(headings: List[str], display_rows: List[Tuple]):
    table = Displayable.from_table(headings=headings, rows=display_rows, max_width=0, center_align=True)
    printer.safe_print(table)


def plural(x, singular, plural):
    if x == 1:
        return singular
    else:
        return plural


def no_color_convention() -> bool:
    """Follow convention for ANSI coloring of CLI tools. See no-color.org."""
    for key, value in os.environ.items():
        if key == "NO_COLOR" and value != "":
            return True
    return False


def py_path_to_module(path: Path, repo_root: Path) -> str:
    return str(path.relative_to(repo_root))[: -len(".py")].replace("./", "").replace("/", ".").replace("\\", ".")


def check_version():
    try:
        response = metadata_service.instance().Nop(request=empty_pb2.Empty())
        client_version_msg_info = response._headers().get(_CLIENT_VERSION_INFO_RESPONSE_HEADER)
        client_version_msg_warning = response._headers().get(_CLIENT_VERSION_WARNING_RESPONSE_HEADER)

        # Currently, only _CLIENT_VERSION_INFO_RESPONSE_HEADER and _CLIENT_VERSION_WARNING_RESPONSE_HEADER
        # metadata is used in the response, whose values have str type.
        # The returned types have 3 cases as of PR #3696:
        # - Metadata value type is List[str] if it's returned from go proxy if direct http is used.
        # - Metadata value is first str in List[str] returned from go proxy if grpc gateway is used.
        # - Metadata value type is str if direct grpc is used.
        # The default values of keys that don't exist are empty strings in any of the 3 cases.
        if client_version_msg_info:
            print_version_msg(client_version_msg_info)
        if client_version_msg_warning:
            print_version_msg(client_version_msg_warning, is_warning=True)
    except Exception as e:
        printer.safe_print("Error connecting to tecton server: ", e, file=sys.stderr)
        sys.exit(1)


def click_exception_wrapper(func):
    """
    Decorator for click commands so that non-Click exceptions are re-raised as ClickExceptions,
    which are displayed gracefully and suppress a stack trace by the click library.

    NOTE: This decorator should be included after Click decorators to ensure exceptions are
    raised as ClickExceptions.

    Example:
    -------
    @click.command()
    @click_exception_wrapper
    def my_command():
        pass

    :param func: function that this decorator wraps
    :return: wrapped function that handles exceptions gracefully
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except (
            TectonAPIInaccessibleError,
            TectonAPIValidationError,
            FailedPreconditionError,
            TectonNotFoundError,
            TectonNotImplementedError,
            TectonInternalError,
            TectonOperationCancelledError,
            TectonDeadlineExceededError,
            TectonResourceExhaustedError,
            TectonAbortedError,
            TectonAlreadyExistsError,
            PermissionError,
        ) as e:
            raise click.ClickException(e)

    return wrapper



from datetime import datetime
from typing import List
from typing import Optional


class TectonValidationError(ValueError):
    """
    Exception that indicates a problem in validating user inputs against
    the data in the system. Typically recoverable by the user.
    """

    def __init__(self, message, can_drop_traceback=False):
        """
        Traceback of this exception can be potentially dropped by sdk decorators
        if flag is set accordingly.
        """
        super().__init__(message)
        self.can_drop_traceback = can_drop_traceback


class AccessError(ValueError):
    """
    Exception that indicates a problem in accessing raw data. Information about connecting to data sources can be found here:
    https://docs.tecton.ai/docs/setting-up-tecton/connecting-data-sources
    """


class TectonInternalError(RuntimeError):
    """
    Exception that indicates an unexpected error within Tecton.
    Can be persistent or transient. Recovery typically requires involving
    Tecton support.
    """


class InvalidDatabricksTokenError(Exception):
    """
    Exception that indicates user's databricks token is invalid.
    """


class TectonSnowflakeNotImplementedError(NotImplementedError):
    """
    Exception that indicates a feature is not yet implemented with Snowflake compute.
    """


class TectonAPIValidationError(ValueError):
    """
    Exception that indicates a problem in validating user inputs against
    the data in the system. Typically recoverable by the user.
    """


class TectonNotFoundError(Exception):
    """
    Exception that indicates that the user's request cannot be found in the system.
    """


class TectonAPIInaccessibleError(Exception):
    """
    Exception that indicates a problem connecting to Tecton cluster.
    """


class FailedPreconditionError(Exception):
    """
    Exception that indicates some prequisite has not been met (e.g the CLI/SDK needs to be updated).
    """


class FailedDependencyDownloadError(Exception):
    """
    Exception that indicates a failure during dependency download(s) while creating an environment.
    """


class TectonNotImplementedError(NotImplementedError):
    """
    Exception that a feature is not implemented or available in Tecton.
    """


class TectonOperationCancelledError(Exception):
    """
    Exception that indicates an operation was cancelled, typically by the user.
    """


class TectonDeadlineExceededError(Exception):
    """
    Exception that indicates a deadline was exceeded.
    """


class TectonResourceExhaustedError(Exception):
    """
    Exception that indicates a resource was exhausted.
    """


class TectonAbortedError(Exception):
    """
    Exception that indicates an operation was aborted.
    """


class TectonAlreadyExistsError(Exception):
    """
    Exception that indicates a resource already exists.
    """


def SCHEMA_VALIDATION_INVALID_COLUMNS(
    actual_columns: List[str],
    expected_columns: List[str],
    missing_columns: List[str],
    extraneous_columns: List[str],
) -> TectonValidationError:
    msg = f"Dataframe schema does not match expected schema. Got columns: {', '.join(actual_columns)}. Expected columns: {', '.join(expected_columns)}."
    if missing_columns:
        msg += f" (Missing columns: {', '.join(missing_columns)}.)"
    if extraneous_columns:
        msg += f" (Extraneous columns: {', '.join(extraneous_columns)}.)"
    return TectonValidationError(msg)


def SCHEMA_VALIDATION_COLUMN_TYPE_MISMATCH_ERROR(
    column_name: str, expected_type: str, actual_type: str
) -> TectonValidationError:
    return TectonValidationError(
        f"The schema does not match the resulting dataframe schema. There is a column type mismatch for column '{column_name}', expected {expected_type}, got {actual_type}"
    )


def UDF_ERROR(error: Exception, feature_definition_name: Optional[str] = None) -> TectonValidationError:
    feature_message = "feature definition"
    if feature_definition_name is not None:
        feature_message += f" [{feature_definition_name}]"
    return TectonValidationError(
        f"UDF Error: please review and ensure the correctness of your {feature_message} and the input data passed in. Otherwise please contact Tecton Support for assistance."
        + f" Running the transformation resulted in the following error: {type(error).__name__}: {str(error)} "
    )


def UDF_TYPE_ERROR(error: Exception) -> TectonValidationError:
    return TectonValidationError(
        "UDF Type Error: please ensure that your UDF correctly handles the typing of row values. Make sure to cast dataframe values to the correct type and ensure that you are handling"
        + f" null column values correctly in your UDF. Running the transformation resulted in the following error: {type(error).__name__}: {str(error)} "
    )


def MISSING_SNOWFAKE_CONNECTION_REQUIREMENTS(param: str) -> TectonValidationError:
    return TectonValidationError(
        f"Snowflake connection is missing the variable {param}. Please ensure the following parameters are set when creating your snowflake connection:  database, warehouse, and schema. "
    )


# Snowflake only
def INVALID_SPINE_SQL(error: Exception) -> TectonValidationError:
    return TectonValidationError(
        f"Invalid SQL: please review your SQL for the spine passed in. Received error: {type(error).__name__}: {str(error)} "
    )


def START_TIME_NOT_BEFORE_END_TIME(start_time: datetime, end_time: datetime) -> TectonValidationError:
    return TectonValidationError(f"start_time ({start_time}) must be less than end_time ({end_time}).")


def TIME_RANGE_NOT_BEFORE_FEATURE_START(
    start_time: datetime, end_time: datetime, feature_start_time: datetime
) -> TectonValidationError:
    return TectonValidationError(
        f"Time range ({start_time} - {end_time}) is before feature start time ({feature_start_time})."
    )


def DS_ARGS_MISSING_FIELD(ds_type: str, field: str) -> TectonValidationError:
    return TectonValidationError(f"{ds_type} data source args must contain field {field}.")


REDSHIFT_DS_EITHER_TABLE_OR_QUERY = TectonValidationError(
    "Redshift data source must contain either table or query, but not both."
)


REDSHIFT_DS_MISSING_SPARK_TEMP_DIR = TectonValidationError(
    'Cannot use a locally defined Redshift data source without a tempdir, e.g. spark.read.format("redshift").option("tempdir", "s3a:///"). See the tempdir that Tecton should use via tecton.conf.set("SPARK_REDSHIFT_TEMP_DIR", <your path>) or by setting the SPARK_REDSHIFT_TEMP_DIR= as an environment variable.'
)


class TectonAthenaValidationError(TectonValidationError):
    """
    Exception that indicates a ValidationError with Athena.
    """


class TectonAthenaNotImplementedError(NotImplementedError):
    """
    Exception that indicates a feature is not yet implemented with Athena compute.
    """


FV_BFC_SINGLE_FROM_SOURCE = TectonValidationError(
    "Computing features from source is not supported for Batch Feature Views with incremental_backfills set to True. "
    + "Enable offline materialization for this feature view in a live workspace to use `get_historical_features()`. Alternatively, use `run()` to test this feature view without materializing data."
)


def FV_NEEDS_TO_BE_MATERIALIZED(fv_name: str) -> TectonValidationError:
    return TectonValidationError(
        f"Feature View '{fv_name}' has not been configured for materialization. "
        + "Please use from_source=True when getting features or "
        + "configure offline materialization for this Feature View in a live workspace."
    )


FT_DF_TOO_LARGE = TectonValidationError(
    "Dataframe too large for a single ingestion, consider splitting into smaller ones"
)


def FT_UPLOAD_FAILED(reason: str) -> TectonValidationError:
    return TectonValidationError(f"Failed to upload dataframe: {reason}")


SNOWFLAKE_CONNECTION_NOT_SET = TectonValidationError(
    "Snowflake connection not configured. Please set Snowflake connection using tecton.snowflake_context.set_connection(connection). https://docs.tecton.ai/docs/setting-up-tecton/connecting-to-a-data-platform/tecton-on-snowflake/connecting-notebooks-to-snowflake"
)


def INTERNAL_ERROR_FROM_MDS(message: str, trace_id: Optional[str] = None) -> TectonInternalError:
    suffix = f", trace ID: {trace_id}" if trace_id else ""
    return TectonInternalError(
        f"Internal Tecton server error, please contact support with error details: {message}{suffix}"
    )



import dataclasses
import logging
import os
import re
import time
import typing
from typing import Dict
from typing import Iterable
from typing import Optional
from typing import Union

import attrs


try:
    import duckdb
except ImportError:
    msg = (
        "Couldn't initialize Rift compute. "
        "To use Rift install all Rift dependencies first by executing `pip install tecton[rift]`."
    )
    raise RuntimeError(msg)
import pandas
import pyarrow
import pyarrow.dataset
import pyarrow.fs
import pyarrow.json
import sqlparse
from deltalake import DeltaTable
from duckdb import DuckDBPyConnection

from tecton_core import conf
from tecton_core import id_helper
from tecton_core.duckdb_context import DuckDBContext
from tecton_core.errors import TectonValidationError
from tecton_core.offline_store import BotoOfflineStoreOptionsProvider
from tecton_core.offline_store import JoinKeyBoundaries
from tecton_core.offline_store import OfflineStoreOptionsProvider
from tecton_core.offline_store import S3Options
from tecton_core.query.dialect import Dialect
from tecton_core.query.errors import SQLCompilationError
from tecton_core.query.errors import UserCodeError
from tecton_core.query.nodes import DataSourceScanNode
from tecton_core.query.query_tree_compute import ComputeMonitor
from tecton_core.query.query_tree_compute import SQLCompute
from tecton_core.schema import Schema
from tecton_core.schema_validation import CastError
from tecton_core.schema_validation import tecton_schema_to_arrow_schema
from tecton_core.secrets import SecretResolver
from tecton_core.specs import FileSourceSpec
from tecton_core.specs import PushTableSourceSpec
from tecton_core.time_utils import get_timezone_aware_datetime
from tecton_proto.data import batch_data_source__client_pb2 as batch_data_source_pb2


@dataclasses.dataclass
class _Cause:
    type_name: str
    message: str


_input_error_pattern = re.compile(
    r"Invalid Input Error: arrow_scan: get_next failed\(\): "
    + r"(?:Unknown error|Invalid): (.*)\. Detail: Python exception: (.*)",
    re.DOTALL,
)


def extract_input_error_cause(e: duckdb.InvalidInputException) -> Optional[_Cause]:
    m = _input_error_pattern.match(str(e))
    if m:
        return _Cause(message=m.group(1), type_name=m.group(2))
    else:
        return None


ARROW_TYPE_TO_DUCKDB_TYPE = {
    pyarrow.int32(): duckdb.dtype("INTEGER"),
    pyarrow.int64(): duckdb.dtype("LONG"),
    pyarrow.float32(): duckdb.dtype("FLOAT"),
    pyarrow.float64(): duckdb.dtype("DOUBLE"),
    pyarrow.string(): duckdb.dtype("STRING"),
    pyarrow.bool_(): duckdb.dtype("BOOLEAN"),
    pyarrow.timestamp("ns"): duckdb.dtype("TIMESTAMP"),
    pyarrow.timestamp("us"): duckdb.dtype("TIMESTAMP"),
    pyarrow.timestamp("ms"): duckdb.dtype("TIMESTAMP"),
    pyarrow.timestamp("ns", "UTC"): duckdb.dtype("TIMESTAMPTZ"),
    pyarrow.timestamp("us", "UTC"): duckdb.dtype("TIMESTAMPTZ"),
    pyarrow.timestamp("ms", "UTC"): duckdb.dtype("TIMESTAMPTZ"),
    pyarrow.date32(): duckdb.dtype("DATE"),
    pyarrow.date64(): duckdb.dtype("DATE"),
}


def _arrow_type_to_duckdb_type(arrow_type: pyarrow.DataType) -> duckdb.duckdb.typing.DuckDBPyType:
    if isinstance(arrow_type, pyarrow.ListType):
        return duckdb.list_type(_arrow_type_to_duckdb_type(arrow_type.value_type))

    if isinstance(arrow_type, pyarrow.MapType):
        return duckdb.map_type(
            _arrow_type_to_duckdb_type(arrow_type.key_type),
            _arrow_type_to_duckdb_type(arrow_type.value_type),
        )

    if isinstance(arrow_type, pyarrow.StructType):
        return duckdb.struct_type({field.name: _arrow_type_to_duckdb_type(field.type) for field in arrow_type})

    return ARROW_TYPE_TO_DUCKDB_TYPE[arrow_type]


@attrs.define
class DuckDBCompute(SQLCompute):
    session: "DuckDBPyConnection"
    is_debug: bool = attrs.field(init=False)
    created_views: typing.List[str] = attrs.field(init=False)
    offline_store_options: Iterable[OfflineStoreOptionsProvider] = ()

    @staticmethod
    def from_context(offline_store_options: Iterable[OfflineStoreOptionsProvider] = ()) -> "DuckDBCompute":
        return DuckDBCompute(
            session=DuckDBContext.get_instance().get_connection(), offline_store_options=offline_store_options
        )

    def __attrs_post_init__(self):
        self.is_debug = conf.get_bool("DUCKDB_DEBUG")
        self.created_views = []

    def run_sql(
        self,
        sql_string: str,
        return_dataframe: bool = False,
        expected_output_schema: Optional[Schema] = None,
        monitor: Optional[ComputeMonitor] = None,
    ) -> Optional[pyarrow.RecordBatchReader]:
        # Notes on case sensitivity:
        # 1. DuckDB is case insensitive when referring to column names, though preserves the
        #    underlying data casing when exporting to e.g. parquet.
        #    See https://duckdb.org/2022/05/04/friendlier-sql.html#case-insensitivity-while-maintaining-case
        #    This means that when using Snowflake for pipeline compute, the view + m13n schema is auto upper-cased
        # 2. When there is a spine provided, the original casing of that spine is used (since DuckDB separately
        #    registers the spine).
        # 3. When exporting values out of DuckDB (to user, or for ODFVs), we coerce the casing to respect the
        #    explicit schema specified. Thus ODFV definitions should reference the casing specified in the dependent
        #    FV's m13n schema.
        sql_string = sqlparse.format(sql_string, reindent=True)
        if self.is_debug:
            logging.warning(f"DUCKDB: run SQL {sql_string}")

        if monitor:
            monitor.set_query(sql_string)

        # Need to use DuckDB cursor (which creates a new connection based on the original connection)
        # to be thread-safe. It avoids a mysterious "unsuccessful or closed pending query result" error too.
        try:
            cursor = self.session.cursor()
            # Although we set timezone globally, DuckDB still needs this cursor-level config to produce
            # correct arrow result. Otherwise, timestamps in arrow table will have a local timezone.
            cursor.sql("SET TimeZone='UTC'")
            duckdb_relation = cursor.sql(sql_string)
            if return_dataframe:
                # increased batch_size (default value 1_000_000) should help Delta writer
                # to produce less fragmented files when data is not sorted by timestamp
                res = duckdb_relation.fetch_arrow_reader(batch_size=10_000_000)
            else:
                res = None

            return res
        except duckdb.InvalidInputException as e:
            # This means that the iterator we passed into DuckDB failed. If it failed due a TectonValidationError
            # we want to unwrap that to get rid of the noisy DuckDB context which is generally irrelevant to the
            # failure.
            cause = extract_input_error_cause(e)
            if not cause:
                raise
            for error_t in (CastError, TectonValidationError):
                if cause.type_name == error_t.__name__:
                    raise error_t(cause.message) from None
            raise
        except duckdb.Error as e:
            raise SQLCompilationError(str(e), sql_string) from None

        return None

    def get_dialect(self) -> Dialect:
        return Dialect.DUCKDB

    def register_temp_table_from_pandas(self, table_name: str, pandas_df: pandas.DataFrame) -> None:
        self.session.from_df(pandas_df).create_view(table_name)
        self.created_views.append(table_name)

    def register_temp_table(
        self, table_name: str, table_or_reader: Union[pyarrow.Table, pyarrow.RecordBatchReader]
    ) -> None:
        self.session.from_arrow(table_or_reader).create_view(table_name)
        self.created_views.append(table_name)

    def register_temp_table_from_offline_store(
        self,
        table_name: str,
        dataset: pyarrow.dataset.Dataset,
        dataset_root_path: str,
        join_keys_filter: Optional[Dict[str, JoinKeyBoundaries]] = None,
        s3_options: Optional[S3Options] = None,
    ) -> None:
        """
        Read an offline store (Delta or Parquet) using DuckDB's Parquet reader.
        Although this function receives pyarrow.Dataset, we don't actually use pyarrow to read the data.
        `pyarrow.Dataset` is just a container with a list of files (fragments) and a schema.
        Pyarrow's schema is converted into DuckDB types to verify that loaded data is consistent with expected schema.

        We use Python wrappers around DuckDB's relational API (https://duckdb.org/docs/api/python/relational_api)
        instead of SQL query, because it's easier to pass join keys filters this way.

        We also need to register S3 credentials in DuckDB to authenticate its Parquet reader.
        See docs for secrets API https://duckdb.org/docs/extensions/httpfs/s3api
        """
        # TODO (vitaly): will have parity with AWS setup pending support from DuckDB around `read_parquet` from GCS
        # For now we'll use the arrow reader to read from GCS.
        if dataset_root_path.startswith("gs"):
            relation = self.session.from_arrow(dataset)
            relation.create_view(table_name)
            return

        files = [os.path.join(dataset_root_path, fragment.path) for fragment in dataset.get_fragments()]
        if not files:
            self.session.from_arrow(dataset.schema.empty_table()).create_view(table_name)
            return

        if s3_options:
            assert s3_options.region, "AWS Region must be specified. Consider setting CLUSTER_REGION in tecton.conf"
            self.run_sql(
                f"""CREATE OR REPLACE secret {table_name} (
                type s3,
                key_id '{s3_options.access_key_id}',
                secret '{s3_options.secret_access_key}',
                session_token '{s3_options.session_token}',
                scope '{dataset_root_path}',
                endpoint 's3.{s3_options.region}.amazonaws.com',
                region '{s3_options.region}')"""
            )

        relation = self.session.read_parquet(files, hive_partitioning=True, union_by_name=True)
        relation = relation.select(
            *[
                duckdb.ColumnExpression(field.name).cast(_arrow_type_to_duckdb_type(field.type)).alias(field.name)
                for field in dataset.schema
            ]
        )

        if join_keys_filter:
            for join_key, boundaries in join_keys_filter.items():
                relation = relation.filter(
                    (duckdb.ColumnExpression(join_key) >= _duckdb_value(boundaries.min))
                    & (duckdb.ColumnExpression(join_key) <= _duckdb_value(boundaries.max))
                )

        relation.create_view(table_name)

    def register_temp_table_from_data_source(
        self,
        table_name: str,
        ds: DataSourceScanNode,
        secret_resolver: Optional[SecretResolver] = None,
        monitor: Optional[ComputeMonitor] = None,
    ) -> None:
        assert isinstance(
            ds.ds.batch_source,
            (
                FileSourceSpec,
                PushTableSourceSpec,
            ),
        ), "DuckDB compute supports only File and Push Table data sources"
        if isinstance(ds.ds.batch_source, FileSourceSpec):
            batch_source_spec = ds.ds.batch_source
            file_uri = batch_source_spec.uri
            timestamp_field = batch_source_spec.timestamp_field

            # ToDo: log loading progress via ComputeMonitor
            schema = Schema(ds.ds.schema.tecton_schema) if ds.ds.schema else None
            arrow_schema = tecton_schema_to_arrow_schema(schema) if schema else None
            if batch_source_spec.timestamp_format:
                # replace timestamp column type with string,
                # we will convert timestamp with DuckDB (see below)
                timestamp_pos = arrow_schema.names.index(timestamp_field)
                arrow_schema = arrow_schema.set(timestamp_pos, pyarrow.field(timestamp_field, pyarrow.string()))

            proto_format = batch_source_spec.file_format
            if proto_format == batch_data_source_pb2.FILE_DATA_SOURCE_FORMAT_CSV:
                arrow_format = "csv"
            elif proto_format == batch_data_source_pb2.FILE_DATA_SOURCE_FORMAT_JSON:
                arrow_format = "json"
            elif proto_format == batch_data_source_pb2.FILE_DATA_SOURCE_FORMAT_PARQUET:
                arrow_format = "parquet"
            else:
                raise ValueError(batch_data_source_pb2.FileDataSourceFormat.Name(batch_source_spec.file_format))

            fs, path = pyarrow.fs.FileSystem.from_uri(file_uri)
            if isinstance(fs, pyarrow.fs.S3FileSystem):
                options = BotoOfflineStoreOptionsProvider.static_options()
                if options is not None:
                    fs = pyarrow.fs.S3FileSystem(
                        access_key=options.access_key_id,
                        secret_key=options.secret_access_key,
                        session_token=options.session_token,
                        # When created via Filesystem.from_uri, the bucket region will be autodetected. This constructor
                        # does not have a bucket from which it can detect the region, so we need to copy it over from the
                        # previous instance.
                        region=fs.region,
                    )

            # There seems to be a bug in Arrow related to the explicit schema:
            # when we pass an explicit schema to `dataset` and both resolution and timezone in the timestamp column
            # don't match the schema in parquet files - filters that are pushed down by DuckDB will not work.
            # It is very likely that we will not guess both resolution and timezone correctly.
            # So we won't pass schema for now.
            arrow_schema = arrow_schema if arrow_format != "parquet" else None
            file_ds = pyarrow.dataset.dataset(source=path, schema=arrow_schema, filesystem=fs, format=arrow_format)
            if batch_source_spec.post_processor:
                reader = pyarrow.RecordBatchReader.from_batches(file_ds.schema, file_ds.to_batches())
                input_df = reader.read_pandas()
                try:
                    processed_df = batch_source_spec.post_processor(input_df)
                except Exception as exc:
                    msg = "Post processor function of data source " f"('{ds.ds.name}') " f"failed with exception"
                    raise UserCodeError(msg) from exc
                else:
                    relation = self.session.from_df(processed_df)
            else:
                relation = self.session.from_arrow(file_ds)

            column_types = dict(zip(relation.columns, relation.dtypes))

            if ds.start_time:
                if column_types[timestamp_field] == duckdb.typing.TIMESTAMP_TZ:
                    start_time = get_timezone_aware_datetime(ds.start_time)
                else:
                    start_time = ds.start_time.replace(tzinfo=None)
                relation = relation.filter(f"\"{timestamp_field}\" >= '{start_time}'")
            if ds.end_time:
                if column_types[timestamp_field] == duckdb.typing.TIMESTAMP_TZ:
                    end_time = get_timezone_aware_datetime(ds.end_time)
                else:
                    end_time = ds.end_time.replace(tzinfo=None)
                relation = relation.filter(f"\"{timestamp_field}\" < '{end_time}'")

            if batch_source_spec.timestamp_format:
                conversion_exp = f"strptime(\"{timestamp_field}\", '{batch_source_spec.timestamp_format}')"
                relation = relation.select(f'* REPLACE({conversion_exp} AS "{timestamp_field}")')

        elif isinstance(ds.ds.batch_source, PushTableSourceSpec):
            ds_id = id_helper.IdHelper.from_string(ds.ds.id)
            creds = next(
                filter(
                    lambda o: o is not None,
                    (p.get_s3_options_for_data_source(ds_id) for p in self.offline_store_options),
                ),
                None,
            )
            if not creds:
                msg = f"Unable to retrieve S3 store credentials for data source {ds.ds.name}"
                raise Exception(msg)
            storage_options = {
                "AWS_ACCESS_KEY_ID": creds.access_key_id,
                "AWS_SECRET_ACCESS_KEY": creds.secret_access_key,
                "AWS_SESSION_TOKEN": creds.session_token,
                "AWS_S3_LOCKING_PROVIDER": "dynamodb",
                "AWS_REGION": conf.get_or_raise("CLUSTER_REGION"),
            }
            saved_error = None
            for _ in range(20):
                try:
                    table = DeltaTable(
                        table_uri=ds.ds.batch_source.ingested_data_location, storage_options=storage_options
                    )
                    break
                except OSError as e:
                    saved_error = e
                    time.sleep(0.1)
            else:
                msg = "Failed to read from S3"
                raise TimeoutError(msg) from saved_error
            df = table.to_pyarrow_dataset()
            relation = self.session.from_arrow(df)
        else:
            msg = "DuckDB compute supports only File data sources and Push Table data sources"
            raise Exception(msg)

        relation.create_view(table_name)
        self.created_views.append(table_name)

    def load_table(self, table_name: str, expected_output_schema: Optional[Schema] = None) -> pyarrow.RecordBatchReader:
        return self.run_sql(
            f"select * from {table_name}", return_dataframe=True, expected_output_schema=expected_output_schema
        )

    def cleanup_temp_tables(self):
        for view in self.created_views:
            self.session.unregister(view)
        self.created_views = []


def _duckdb_value(v: Union[str, int, float]) -> duckdb.Value:
    if isinstance(v, str):
        return duckdb.StringValue(v)

    if isinstance(v, int):
        return duckdb.IntegerValue(v)

    if isinstance(v, float):
        return duckdb.DoubleValue(v)

    msg = f"Unsupported value type {type(v)} for value: {v}"
    raise TypeError(msg)



import re
from typing import Callable
from typing import Optional
from typing import Tuple

from google.api_core.exceptions import PreconditionFailed
from google.cloud import storage

from tecton_core.id_helper import IdHelper
from tecton_materialization.common.job_metadata import JobMetadataClient
from tecton_proto.common.id__client_pb2 import Id
from tecton_proto.materialization.job_metadata__client_pb2 import JobMetadata
from tecton_proto.materialization.params__client_pb2 import MaterializationTaskParams


class GCSMetadataClient(JobMetadataClient):
    def __init__(self, table: str, attempt_id: Id):
        matches = re.match("gs://(.*?)/(.*)", f"{table}/{IdHelper.to_string(attempt_id)}")
        self._bucket_name, self._blob_name = matches.groups()
        self._storage_client = storage.Client()

    @staticmethod
    def for_params(materialization_task_params: MaterializationTaskParams) -> JobMetadataClient:
        return GCSMetadataClient(materialization_task_params.job_metadata_table, materialization_task_params.attempt_id)

    def get(self) -> Tuple[JobMetadata, int]:
        bucket = self._storage_client.bucket(self._bucket_name)
        blob = bucket.blob(self._blob_name)
        item = blob.download_as_string()
        data = JobMetadata()
        data.ParseFromString(item)
        assert blob.generation is not None, (self._bucket_name, self._blob_name, blob)
        return data, blob.generation

    def update(self, updater: Callable[[JobMetadata], Optional[JobMetadata]]) -> JobMetadata:
        num_retries = 100
        last_exception = None
        for i in range(num_retries):
            try:
                old_data, old_version = self.get()
                new_data = updater(old_data)
                if new_data is None:
                    return old_data
                bucket = self._storage_client.bucket(self._bucket_name)
                blob = bucket.blob(self._blob_name)
                blob.upload_from_string(new_data.SerializeToString(), if_generation_match=old_version)
                return new_data
            except PreconditionFailed as e:
                # We had a conflicting update
                last_exception = e
                continue
        raise last_exception



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/snowflake/snowflake_credentials__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n:tecton_proto/snowflake/snowflake_credentials__client.proto\x12\x16tecton_proto.snowflake\"\x9f\x01\n\x14SnowflakeCredentials\x12\x12\n\x04user\x18\x01 \x01(\tR\x04user\x12\x32\n\x14password_secret_name\x18\x02 \x01(\tH\x00R\x12passwordSecretName\x12,\n\x11private_key_alias\x18\x03 \x01(\tH\x00R\x0fprivateKeyAliasB\x11\n\x0f\x63redential_typeB\x18\n\x14\x63om.tecton.snowflakeP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.snowflake.snowflake_credentials__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\024com.tecton.snowflakeP\001'
  _SNOWFLAKECREDENTIALS._serialized_start=87
  _SNOWFLAKECREDENTIALS._serialized_end=246
# @@protoc_insertion_point(module_scope)



import logging
import os
import sys

from tecton_core import materialization_context


# TODO(deprecated_after=0.5): handle backward-compatibility for customer copies of builtin transformations that did not use tecton.materialization_context
# but instead directly accessed tecton_spark.materialization_context
sys.modules["tecton_spark.materialization_context"] = materialization_context

if os.environ.get("PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION", "") == "cpp":
    msg = (
        "The 'cpp' implementation of protobuf cannot be made to work in a federated environment. "
        "The 'upb' implementation is about as fast and will almost certainly work; it is usually the default."
    )
    raise AssertionError(msg)

try:
    logging.getLogger("py4j").setLevel(logging.WARN)
except Exception:
    pass



import tempfile
from unittest import TestCase

from tecton._internals.spark_api import _get_checkpoint_dir_name


class SparkApiTest(TestCase):
    def test_get_checkpoint_dir(self):
        tmp_dir = tempfile.mkdtemp(prefix="abc_")
        checkpoint_dir = _get_checkpoint_dir_name(tmp_dir)
        assert checkpoint_dir.startswith(tmp_dir)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/amplitude/client_logging__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n3tecton_proto/amplitude/client_logging__client.proto\x12\x16tecton_proto.amplitude\x1a\x1egoogle/protobuf/duration.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"7\n\x0bLoggedValue\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value\"|\n\x08\x45rrorLog\x12\x18\n\x07message\x18\x01 \x01(\tR\x07message\x12\x1e\n\nstacktrace\x18\x02 \x01(\tR\nstacktrace\x12\x36\n\x05\x63\x61use\x18\x03 \x01(\x0b\x32 .tecton_proto.amplitude.ErrorLogR\x05\x63\x61use\"\xf0\x04\n\x13SDKMethodInvocation\x12.\n\x04time\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x04time\x12\x17\n\x07user_id\x18\x02 \x01(\tR\x06userId\x12\x19\n\x08trace_id\x18\x03 \x01(\tR\x07traceId\x12\x1b\n\tlog_level\x18\x04 \x01(\tR\x08logLevel\x12:\n\x04type\x18\x05 \x01(\x0e\x32&.tecton_proto.amplitude.LogMessageTypeR\x04type\x12\x1d\n\nclass_name\x18\x06 \x01(\tR\tclassName\x12\x1f\n\x0bmethod_name\x18\x07 \x01(\tR\nmethodName\x12@\n\x0e\x65xecution_time\x18\x08 \x01(\x0b\x32\x19.google.protobuf.DurationR\rexecutionTime\x12Z\n\x17params_or_return_values\x18\t \x03(\x0b\x32#.tecton_proto.amplitude.LoggedValueR\x14paramsOrReturnValues\x12\x36\n\x05\x65rror\x18\n \x01(\x0b\x32 .tecton_proto.amplitude.ErrorLogR\x05\x65rror\x12\x1c\n\tworkspace\x18\x0b \x01(\tR\tworkspace\x12\x1f\n\x0bsdk_version\x18\x0c \x01(\tR\nsdkVersion\x12%\n\x0epython_version\x18\r \x01(\tR\rpythonVersion\x12 \n\x0cis_local_fco\x18\x0e \x01(\x08R\nisLocalFco*\x8a\x01\n\x0eLogMessageType\x12#\n\x1f\x43LIENT_LOG_MESSAGE_TYPE_UNKNOWN\x10\x00\x12(\n$CLIENT_LOG_MESSAGE_TYPE_METHOD_ENTRY\x10\x01\x12)\n%CLIENT_LOG_MESSAGE_TYPE_METHOD_RETURN\x10\x02\x42,\n\x14\x63om.tecton.amplitudeB\x12\x43lientLoggingProtoP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.amplitude.client_logging__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\024com.tecton.amplitudeB\022ClientLoggingProtoP\001'
  _LOGMESSAGETYPE._serialized_start=955
  _LOGMESSAGETYPE._serialized_end=1093
  _LOGGEDVALUE._serialized_start=144
  _LOGGEDVALUE._serialized_end=199
  _ERRORLOG._serialized_start=201
  _ERRORLOG._serialized_end=325
  _SDKMETHODINVOCATION._serialized_start=328
  _SDKMETHODINVOCATION._serialized_end=952
# @@protoc_insertion_point(module_scope)



from datetime import datetime
from typing import Any
from typing import Dict
from typing import List

from pyspark.sql import DataFrame
from pyspark.sql import Row
from pyspark.sql import types as spark_type

from tecton_core import errors
from tecton_core.schema import Schema
from tecton_spark.schema_spark_utils import schema_from_spark


def validate_df_columns_and_feature_types(
    df: DataFrame, view_schema: Schema, allow_extraneous_columns: bool = True
) -> None:
    df_columns = frozenset(schema_from_spark(df.schema).column_name_and_data_types())
    df_column_names = frozenset([x[0] for x in df_columns])
    fv_columns = view_schema.column_name_and_data_types()
    fv_column_names = frozenset([x[0] for x in fv_columns])

    missing_columns = fv_column_names - df_column_names
    extraneous_columns = df_column_names - fv_column_names

    invalid_column_set = missing_columns or (extraneous_columns and not allow_extraneous_columns)
    if invalid_column_set:
        raise errors.SCHEMA_VALIDATION_INVALID_COLUMNS(
            actual_columns=sorted(df_column_names),
            expected_columns=sorted(fv_column_names),
            extraneous_columns=sorted(extraneous_columns),
            missing_columns=sorted(missing_columns),
        )

    for fv_column in fv_columns:
        if fv_column not in df_columns:
            raise errors.SCHEMA_VALIDATION_COLUMN_TYPE_MISMATCH_ERROR(
                fv_column[0], fv_column[1], next(x for x in df_columns if x[0] == fv_column[0])[1]
            )


PYTHON_TO_SPARK_TYPE_MAPPING = {
    int: (spark_type.IntegerType, spark_type.LongType),
    str: (spark_type.StringType),
    float: (spark_type.FloatType, spark_type.DoubleType),
    bool: (spark_type.BooleanType),
    list: (spark_type.ArrayType),
    dict: (spark_type.MapType),
    datetime: (spark_type.TimestampType),
}


def _is_acceptable_struct_type(
    python_value: Any,  # noqa: ANN401
    expected_struct_type: spark_type.StructType,
    field_path: str,
    odfv_name: str,
) -> None:
    # By default, the input struct value in pyspark udf is represented as `Row` class. Depends on user's transformation,
    # the output type can be a `Row` or a `dict`.
    if isinstance(python_value, Row):
        python_value = python_value.asDict()

    if not isinstance(python_value, dict):
        msg = f"On-Demand Feature View '{odfv_name}' has a field '{'.'.join(field_path)}' that is expected to be a struct, but got {type(python_value).__name__}."
        raise TypeError(msg)

    extraneous_fields = set(python_value.keys()) - set(expected_struct_type.fieldNames())
    if extraneous_fields:
        extraneous = ", ".join(extraneous_fields)
        msg = f"On-Demand Feature View '{odfv_name}' has extraneous fields in '{'.'.join(field_path)}': {extraneous}"
        raise TypeError(msg)

    for sub_field in expected_struct_type.fields:
        field_name = sub_field.name
        field_type = sub_field.dataType
        if field_name in python_value:
            field_path.append(field_name)
            _is_acceptable_type(python_value[field_name], field_type, field_path, odfv_name)
            field_path.pop()


def _is_acceptable_array_type(
    python_value: Any,  # noqa: ANN401
    expected_array_type: spark_type.ArrayType,
    field_path: List[str],
    odfv_name: str,
) -> None:
    if not isinstance(python_value, list):
        msg = f"On-Demand Feature View '{odfv_name}' has a field '{'.'.join(field_path)}' that is expected to be a list, but got {type(python_value).__name__}."
        raise TypeError(msg)

    current_field_name = field_path.pop()
    for i, item in enumerate(python_value):
        field_path.append(f"{current_field_name}[{i}]")
        _is_acceptable_type(item, expected_array_type.elementType, field_path, odfv_name)
        field_path.pop()
    field_path.append(current_field_name)


def _is_acceptable_map_type(
    python_value: Any,  # noqa: ANN401
    expected_map_type: spark_type.MapType,
    field_path: List[str],
    odfv_name: str,
) -> None:
    if not isinstance(python_value, dict):
        msg = f"On-Demand Feature View '{odfv_name}' has a field '{'.'.join(field_path)}' that is expected to be a dict, but got {type(python_value).__name__}."
        raise TypeError(msg)

    map_field_name = field_path.pop()
    for key, value in python_value.items():
        field_path.extend([map_field_name, "key"])
        _is_acceptable_type(key, expected_map_type.keyType, field_path, odfv_name)
        field_path.pop()
        field_path.pop()

        field_path.append(f"{map_field_name}[{key}]")
        _is_acceptable_type(value, expected_map_type.valueType, field_path, odfv_name)
        field_path.pop()

    field_path.append(map_field_name)


def _is_acceptable_type(
    python_value: Any,  # noqa: ANN401
    expected_type: spark_type.DataType,
    field_path: List[str],
    odfv_name: str,
) -> None:
    if python_value is None:
        return

    if isinstance(expected_type, spark_type.StructType):
        _is_acceptable_struct_type(python_value, expected_type, field_path, odfv_name)

    elif isinstance(expected_type, spark_type.ArrayType):
        _is_acceptable_array_type(python_value, expected_type, field_path, odfv_name)

    elif isinstance(expected_type, spark_type.MapType):
        _is_acceptable_map_type(python_value, expected_type, field_path, odfv_name)

    # Check non-complex data types
    else:
        acceptable_types = PYTHON_TO_SPARK_TYPE_MAPPING.get(type(python_value), ())
        if not isinstance(expected_type, acceptable_types):
            msg = f"On-Demand Feature View '{odfv_name}' has a field '{'.'.join(field_path)}' that is expected to be of type {expected_type.simpleString()}, but got type {type(python_value).__name__}."
            raise TypeError(msg)


def check_python_odfv_output_schema(
    output_dict: Dict[str, Any], expected_schema: spark_type.StructType, odfv_name: str
) -> None:
    extraneous_fields = set(output_dict.keys()) - set(expected_schema.fieldNames())
    if extraneous_fields:
        msg = f"On-Demand Feature View '{odfv_name}' has unexpected columns found in the dataframe schema: {', '.join(list(extraneous_fields))}. Expected schema: {', '.join(list(expected_schema.fieldNames()))}"
        raise TypeError(msg)

    for field_name in output_dict:
        output_py_value = output_dict[field_name]
        expected_spark_type = expected_schema[field_name].dataType
        _is_acceptable_type(output_py_value, expected_spark_type, [field_name], odfv_name)



import re

from tecton_core import data_types as tecton_types
from tecton_core import errors
from tecton_proto.common import data_type__client_pb2 as data_type_pb2


# Keep in sync with DataTypeUtils.kt
def snowflake_type_to_tecton_type(snowflake_type: str, column_name: str) -> tecton_types.DataType:
    if snowflake_type.startswith("VARCHAR"):
        return tecton_types.StringType()
    elif snowflake_type.startswith("NUMBER"):
        m = re.fullmatch("NUMBER\\((?P<precision>\\d+),(?P<scale>\\d+)\\)", snowflake_type)
        if m and int(m.groupdict().get("scale", -1)) > 0:
            return tecton_types.Float64Type()
        return tecton_types.Int64Type()
    elif snowflake_type.startswith("FLOAT"):
        return tecton_types.Float64Type()
    elif snowflake_type.startswith("BOOLEAN"):
        return tecton_types.BoolType()
    elif snowflake_type.startswith("TIMESTAMP_NTZ"):
        return tecton_types.TimestampType()
    elif snowflake_type.startswith(("TIMESTAMP_TZ", "TIMESTAMP_LTZ")):
        msg = f"Timestamp type {snowflake_type} for column {column_name} is not supported because it contains a timezone.Please use TIMESTAMP_NTZ instead."
        raise errors.TectonValidationError(msg)
    else:
        msg = f"Unsupported Snowflake type {snowflake_type} for column {column_name}."
        raise errors.TectonValidationError(msg)


def tecton_type_to_snowflake_type(tecton_type_proto: data_type_pb2.DataType, column_name: str) -> str:
    tecton_type = tecton_types.data_type_from_proto(tecton_type_proto)
    if isinstance(tecton_type, tecton_types.StringType):
        return "VARCHAR"
    elif isinstance(tecton_type, tecton_types.Int64Type):
        return "NUMBER"
    elif isinstance(tecton_type, tecton_types.Float64Type):
        return "FLOAT"
    elif isinstance(tecton_type, tecton_types.BoolType):
        return "BOOLEAN"
    elif isinstance(tecton_type, tecton_types.TimestampType):
        return "TIMESTAMP_NTZ"
    elif isinstance(tecton_type, tecton_types.ArrayType) and isinstance(
        tecton_type.element_type, tecton_types.StringType
    ):
        return "ARRAY"
    else:
        msg = f"Unsupported tecton type {tecton_type} for column {column_name} on Snowflake."
        raise errors.TectonValidationError(msg)
